from .Base import Base
from pegasus.tools import run_find_markers


class FindMarkers(Base):
    """
Find markers using gradient boosting.

Usage:
  pegasus find_markers [options] <input_h5ad_file> <output_spreadsheet>
  pegasus find_markers -h

Arguments:
  input_h5ad_file        Single cell data after running the de_analysis.
  output_spreadsheet     Output spreadsheet with LightGBM detected markers.

Options:
  -p <threads>                 Use <threads> threads. [default: 1]
  --labels <attr>              <attr> used as cluster labels. [default: louvain_labels]
  --de-key <key>               Key for storing DE results in 'varm' field. [default: de_res]
  --remove-ribo                Remove ribosomal genes with either RPL or RPS as prefixes.
  --min-gain <gain>            Only report genes with a feature importance score (in gain) of at least <gain>. [default: 1.0]
  --random-state <seed>        Random state for initializing LightGBM and KMeans. [default: 0]

  -h, --help                   Print out help information.

Outputs:
  output_spreadsheet     An excel spreadsheet containing detected markers. Each cluster has one tab in the spreadsheet and each tab has six columns, listing markers that are strongly up-regulated, weakly up-regulated, down-regulated and their associated LightGBM gains.

Examples:
  pegasus find_markers --labels louvain_labels --remove-ribo --min-gain 10.0 -p 10 manton_bm.h5ad manton_bm.markers.xlsx
    """

    def execute(self):
        run_find_markers(
            self.args["<input_h5ad_file>"],
            self.args["<output_spreadsheet>"],
            self.args["--labels"],
            de_key=self.args["--de-key"],
            n_jobs=int(self.args["-p"]),
            min_gain=float(self.args["--min-gain"]),
            random_state=int(self.args["--random-state"]),
            remove_ribo=self.args["--remove-ribo"],
        )



from .Base import Base
from pegasus.tools import aggregate_matrices


class AggregateMatrix(Base):
    """
Aggregate sample count matrices from each channel into one big matrix.

Usage:
  pegasus aggregate_matrix <csv_file> <output_name> [--restriction <restriction>... --attributes <attributes> --default-reference <reference> --select-only-singlets --minimum-number-of-genes <ngene>]
  pegasus aggregate_matrix -h

Arguments:
  csv_file          Input csv-formatted file containing information of each sc/snRNA-seq sample. This file must contain at least 2 columns - Sample, sample name and Location, location of the sample count matrix in either 10x v2/v3, DGE, mtx, csv, tsv or loom format. Additionally, an optional Reference column can be used to select samples generated from a same reference (e.g. mm10). If the count matrix is in either DGE, mtx, csv, tsv, or loom format, the value in this column will be used as the reference since the count matrix file does not contain reference name information. In addition, the Reference column can be used to aggregate count matrices generated from different genome versions or gene annotations together under a unified reference. For example, if we have one matrix generated from mm9 and the other one generated from mm10, we can write mm9_10 for these two matrices in their Reference column. Pegasus will change their references to 'mm9_10' and use the union of gene symbols from the two matrices as the gene symbols of the aggregated matrix. For HDF5 files (e.g. 10x v2/v3), the reference name contained in the file does not need to match the value in this column. In fact, we use this column to rename references in HDF5 files. For example, if we have two HDF files, one generated from mm9 and the other generated from mm10. We can set these two files' Reference column value to 'mm9_10', which will rename their reference names into mm9_10 and the aggregated matrix will contain all genes from either mm9 or mm10. This renaming feature does not work if one HDF5 file contain multiple references (e.g. mm10 and GRCh38).
  output_name       The output file name.

Options:
  --restriction <restriction>...           Select channels that satisfy all restrictions. Each restriction takes the format of name:value,...,value or name:~value,..,value, where ~ refers to not. You can specifiy multiple restrictions by setting this option multiple times.
  --attributes <attributes>                Specify a comma-separated list of outputted attributes. These attributes should be column names in the csv file.
  --default-reference <reference>          If sample count matrix is in either DGE, mtx, csv, tsv or loom format and there is no Reference column in the csv_file, use <reference> as the reference.
  --select-only-singlets                   If we have demultiplexed data, turning on this option will make pegasus only include barcodes that are predicted as singlets.
  --minimum-number-of-genes <ngene>        Only keep barcodes with at least <ngene> expressed genes.
  -h, --help                               Print out help information.

Outputs:
  output_name.h5sc        A pegasus-formatted HDF5 file containing the count matrices and associated attributes.

Examples:
  pegasus aggregate_matrix --restriction Source:BM,CB --restriction Individual:1-8 --attributes Source,Platform Manton_count_matrix.csv manton_bm_cb
    """

    def execute(self):
        aggregate_matrices(
            self.args["<csv_file>"],
            what_to_return=self.args["<output_name>"],
            restrictions=self.args["--restriction"],
            attributes=self.split_string(self.args["--attributes"]),
            default_ref=self.args["--default-reference"],
            select_singlets=self.args["--select-only-singlets"],
            ngene=self.convert_to_int(self.args["--minimum-number-of-genes"]),
        )



import time
import numpy as np
from scipy.sparse import issparse
from anndata import AnnData
import logging

from pegasus.tools import estimate_feature_statistics, select_features, X_from_rep

logger = logging.getLogger("pegasus")
from pegasus.utils import decorators as pg_deco



def set_group_attribute(data: AnnData, attribute_string: str) -> None:
    """Set group attributes used in batch correction.

    Batch correction assumes the differences in gene expression between channels are due to batch effects. However, in many cases, we know that channels can be partitioned into several groups and each group is biologically different from others. In this case, *pegasus* will only perform batch correction for channels within each group.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    attribute_string: ``str``
        Attributes used to construct groups:

        * ``None``
            Assume all channels are from one group.

        * ``attr`` 
            Define groups by sample attribute ``attr``, which is a keyword in ``data.obs``.

        * ``att1+att2+...+attrn`` 
            Define groups by the Cartesian product of these *n* attributes, which are keywords in ``data.obs``.

        * ``attr=value_11,...value_1n_1;value_21,...value_2n_2;...;value_m1,...,value_mn_m``
            In this form, there will be *(m+1)* groups. A cell belongs to group *i* (*i > 1*) if and only if its sample attribute ``attr``, which is a keyword in ``data.obs``, has a value among ``value_i1``, ... ``value_in_i``. A cell belongs to group 0 if it does not belong to any other groups.

    Returns
    -------
    None

        Update ``data.obs``:
        
        * ``data.obs["Group"]``: Group ID for each cell.

    Examples
    --------

    >>> pg.set_group_attribute(adata, attr_string = "Individual")

    >>> pg.set_group_attribute(adata, attr_string = "Individual+assignment")

    >>> pg.set_group_attribute(adata, attr_string = "Channel=1,3,5;2,4,6,8")    
    """
    
    if attribute_string.find("=") >= 0:
        attr, value_str = attribute_string.split("=")
        assert attr in data.obs.columns
        values = value_str.split(";")
        data.obs["Group"] = "0"
        for group_id, value in enumerate(values):
            vals = value.split(",")
            idx = np.isin(data.obs[attr], vals)
            data.obs.loc[idx, "Group"] = str(group_id + 1)
    elif attribute_string.find("+") >= 0:
        attrs = attribute_string.split("+")
        assert np.isin(attrs, data.obs.columns).sum() == len(attrs)
        data.obs["Group"] = data.obs[attrs].apply(lambda x: "+".join(x), axis=1)
    else:
        assert attribute_string in data.obs.columns
        data.obs["Group"] = data.obs[attribute_string]

def estimate_adjustment_matrices(data: AnnData) -> bool:
    """ Estimate adjustment matrices
    """

    if "plus" in data.varm.keys() or "muls" in data.varm.keys():
        # This only happens if this is for subclustering. Thus do not calculate factors, using factors calculated from parent for batch correction.
        assert "plus" in data.varm.keys() and "muls" in data.varm.keys()
        return True

    if ("gmeans" not in data.varm) or ("gstds" not in data.varm):
        estimate_feature_statistics(data, True)

    if data.uns["Channels"].size == 1:
        logger.warning(
            "Warning: data only contains 1 channel. Batch correction disabled!"
        )
        return False

    nchannel = data.uns["Channels"].size

    plus = np.zeros((data.shape[1], nchannel))
    muls = np.zeros((data.shape[1], nchannel))

    ncells = data.uns["ncells"]
    means = data.varm["means"]
    partial_sum = data.varm["partial_sum"]
    gmeans = data.varm["gmeans"]
    gstds = data.varm["gstds"]
    c2gid = data.uns["c2gid"]
    for i in range(data.uns["Channels"].size):
        if ncells[i] > 1:
            muls[:, i] = (partial_sum[:, i] / (ncells[i] - 1.0)) ** 0.5
        outliers = muls[:, i] < 1e-6
        normals = np.logical_not(outliers)
        muls[outliers, i] = 1.0
        muls[normals, i] = gstds[normals, c2gid[i]] / muls[normals, i]
        plus[:, i] = gmeans[:, c2gid[i]] - muls[:, i] * means[:, i]

    data.varm["plus"] = plus
    data.varm["muls"] = muls

    return True

def correct_batch_effects(data: AnnData, keyword: str, features: str = None) -> None:
    """ Apply calculated plus and muls to correct batch effects for a dense matrix
    """
    X = data.uns[keyword]
    m = X.shape[1]
    if features is not None:
        selected = data.var[features].values
        plus = data.varm["plus"][selected, :]
        muls = data.varm["muls"][selected, :]
    else:
        selected = np.ones(data.shape[1], dtype=bool)
        plus = data.varm["plus"]
        muls = data.varm["muls"]

    for i, channel in enumerate(data.uns["Channels"]):
        idx = np.isin(data.obs["Channel"], channel)
        if idx.sum() == 0:
            continue
        X[idx] = X[idx] * np.reshape(muls[:, i], newshape=(1, m)) + np.reshape(
            plus[:, i], newshape=(1, m)
        )
    # X[X < 0.0] = 0.0

def correct_batch(data: AnnData, features: str = None) -> None:
    """Batch correction on data.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    features: `str`, optional, default: ``None``
        Features to be included in batch correction computation. If ``None``, simply consider all features.

    Returns
    -------
    ``None``

    Update ``data.X`` by the corrected count matrix.

    Examples
    --------
    >>> pg.correct_batch(adata, features = "highly_variable_features")
    """

    tot_seconds = 0.0

    # estimate adjustment parameters
    start = time.perf_counter()
    can_correct = estimate_adjustment_matrices(data)
    end = time.perf_counter()
    tot_seconds += end - start
    logger.info("Adjustment parameters are estimated.")

    # select dense matrix
    keyword = select_features(data, features)
    logger.info("Features are selected.")

    if can_correct:
        start = time.perf_counter()
        correct_batch_effects(data, keyword, features)
        end = time.perf_counter()
        tot_seconds += end - start
        logger.info(
            "Batch correction is finished. Time spent = {:.2f}s.".format(tot_seconds)
        )



@pg_deco.TimeLogger()
def run_harmony(
    data: AnnData,
    rep: str = 'pca',
    n_jobs: int = -1,
    n_clusters: int = None,
    random_state: int = 0,
) -> str:
    """Batch correction PCs using Harmony

    Parameters
    ----------
    data: ``anndata.AnnData``.
        Annotated data matrix with rows for cells and columns for genes.

    rep: ``str``, optional, default: ``"pca"``.
        Which representation to use as input of Harmony, default is PCA.

    n_jobs : ``int``, optional, default: ``-1``.
        Number of threads to use for the KMeans clustering used in Harmony. ``-1`` refers to using all available threads.

    n_clusters: ``int``, optional, default: ``None``.
        Number of Harmony clusters. Default is ``None``, which asks Harmony to estimate this number from the data.

    random_state: ``int``, optional, default: ``0``.
        Seed for random number generator

    Returns
    -------
    out_rep: ``str``
        The keyword in ``data.obsm`` referring to the embedding calculated by Harmony algorithm.

    Update ``data.obsm``:
        * ``data.obsm['X_' + out_rep]``: The embedding calculated by Harmony algorithm.

    Examples
    --------
    >>> pg.run_harmony(adata, rep = "pca", n_jobs = 10, random_state = 25)
    """
    try:
        from harmony import harmonize
    except ImportError:
        print("Need harmony! Try 'pip install harmony-pytorch'.")
    
    logger.info("Start integration using Harmony.")    
    out_rep = rep + '_harmony'
    data.obsm['X_' + out_rep] = harmonize(X_from_rep(data, rep), data.obs, 'Channel', n_clusters = n_clusters, n_jobs_kmeans = n_jobs, random_state = random_state)
    return out_rep



from .Base import Base
from pegasus.io import read_input, write_output


class ConvertH5SC(Base):
    """
Convert H5SC file into 10x V2 formatted HDF5 file.

Usage:
  pegasus convert_h5sc <input.h5sc> <output_10x.h5>
  pegasus convert_h5sc -h

Arguments:
  input.h5sc           Input file name.
  output_10x.h5        Output file name.

Options:
  -h, --help           Print out help information.

Outputs:
  output_name.h5sc        A pegasus-formatted HDF5 file containing the count matrices and associated attributes.

Examples:
  pegasus convert_h5sc input.h5sc output_10x.h5
    """

    def execute(self):
      data = read_input(self.args["<input.h5sc>"], return_type = "MemData")
      write_output(data, self.args["<output_10x.h5>"])




import os
from .Base import Base
from pegasus.tools import run_de_analysis


class DeAnalysis(Base):
    """
Perform DE analysis.

Usage:
  pegasus de_analysis [options] <input_h5ad_file> <output_spreadsheet>
  pegasus de_analysis -h

Arguments:
  input_h5ad_file        Single cell data with clustering calculated. DE results would be written back.
  output_spreadsheet     Output spreadsheet with DE results.

Options:
  -p <threads>                     Use <threads> threads. [default: 1]
  --labels <attr>                  <attr> used as cluster labels. [default: louvain_labels]
  --result-key <key>               Store DE results into AnnData varm with key = <key>. [default: de_res]
  --auc                            Calculate area under ROC (AUROC) and area under Precision-Recall (AUPR).
  --t                              Calculate Welch's t-test.
  --fisher                         Calculate Fisher's exact test.
  --mwu                            Calculate Mann-Whitney U test.
  --temp-folder <temp_folder>      Joblib temporary folder for memmapping numpy arrays.
  --alpha <alpha>                  Control false discovery rate at <alpha>. [default: 0.05]
  --ndigits <ndigits>              Round non p-values and q-values to <ndigits> after decimal point in the excel. [default: 3]

  --quiet                          Do not show detailed intermediate outputs.
  -h, --help                       Print out help information.

Outputs:
  input_h5ad_file        DE results would be written back to the 'varm' field with name set by --result-key <key>.
  output_spreadsheet     An excel spreadsheet containing DE results. Each cluster has two tabs in the spreadsheet. One is for up-regulated genes and the other is for down-regulated genes.

Examples:
  pegasus de_analysis -p 26 --labels louvain_labels --auc --t --fisher --mwu manton_bm.h5ad manton_bm_de.xlsx
    """

    def execute(self):
        run_de_analysis(
            self.args["<input_h5ad_file>"],
            self.args["<output_spreadsheet>"],
            self.args["--labels"],
            result_key=self.args["--result-key"],
            n_jobs=int(self.args["-p"]),
            auc=self.args["--auc"],
            t=self.args["--t"],
            fisher=self.args["--fisher"],
            mwu=self.args["--mwu"],
            temp_folder=self.args["--temp-folder"],
            verbose=not self.args["--quiet"],
            alpha=float(self.args["--alpha"]),
            ndigits=int(self.args["--ndigits"]),
        )



try:
    get_ipython
except NameError:
    import matplotlib

    matplotlib.use("Agg")

import sys
import logging
import warnings

logger = logging.getLogger("pegasus")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

warnings.filterwarnings("ignore", category=UserWarning,  module='lightgbm')

from .io import infer_file_format, read_input, write_output
from .tools import (
    aggregate_matrices,
    qc_metrics,
    get_filter_stats,
    filter_data,
    log_norm,
    select_features,
    pca,
    highly_variable_features,
    set_group_attribute,
    correct_batch,
    run_harmony,
    neighbors,
    calc_kBET,
    calc_kSIM,
    diffmap,
    reduce_diffmap_to_3d,
    calc_pseudotime,
    cluster,
    louvain,
    leiden,
    spectral_louvain,
    spectral_leiden,
    tsne,
    fitsne,
    umap,
    fle,
    net_tsne,
    net_fitsne,
    net_umap,
    net_fle,
    de_analysis,
    markers,
    write_results_to_excel,
    find_markers,
    infer_path,
)
from .annotate_cluster import infer_cell_types, annotate, infer_cluster_names
from .demuxEM import estimate_background_probs, demultiplex
from .misc import search_genes, search_de_genes

from scplot import (
    violin,
    heatmap,
    scatter,
    line,
    dotplot,
    scatter_matrix,
    embedding,
    composition_plot,
    variable_feature_plot,
    volcano,
)

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import numpy as np
from scipy.sparse import issparse


def update_rep(rep: str) -> str:
    """ If rep is None, return rep as mat, which refers to the whole expression matrix
    """
    return rep if rep is not None else "mat"


def X_from_rep(data: "AnnData", rep: str) -> np.array:
    """
    If rep is not mat, first check if X_rep is in data.obsm. If not, raise an error. 
    If rep is None, return data.X as a numpy array
    """
    if rep != "mat":
        rep_key = "X_" + rep
        if rep_key not in data.obsm.keys():
            raise ValueError("Cannot find {0} matrix. Please run {0} first".format(rep))
        return data.obsm[rep_key]
    else:
        return data.X if not issparse(data.X) else data.X.toarray()


def W_from_rep(data: "AnnData", rep: str) -> "csr_matrix":
    """
    Return affinity matrix W based on representation rep.
    """
    rep_key = "W_" + rep
    if rep_key not in data.uns:
        raise ValueError("Affinity matrix does not exist. Please run neighbors first!")
    return data.uns[rep_key]


def knn_is_cached(
    data: "AnnData", indices_key: str, distances_key: str, K: int
) -> bool:
    return (
        (indices_key in data.uns)
        and (distances_key in data.uns)
        and data.uns[indices_key].shape[0] == data.shape[0]
        and (K <= data.uns[indices_key].shape[1] + 1)
    )



import os
from .Base import Base
from pegasus.plotting import make_interactive_plots


class iPlotting(Base):
    """
Generate cluster composition plots.

Usage:
  pegasus iplot --attribute <attr> [options] <basis> <input_h5ad_file> <output_html_file>
  pegasus iplot -h

Arguments:
  basis                  Basis can be either 'tsne', 'fitsne', 'umap', 'diffmap', 'pca', or 'diffmap_pca'.
  input_h5ad_file        Single cell data with clustering done in h5ad file format.
  output_html_file       Output interactive plot in html format.

Options:
  --attribute <attr>        Use attribute <attr> as labels in the plot.
  --is-real                 <attr> is real valued.
  --is-gene                 <attr> is a gene name.
  --log10                   If take log10 of real values.

  -h, --help                Print out help information.

Examples:
  pegasus iplot --attribute louvain_labels tsne Manton_BM.h5ad test.html
  pegasus iplot --attribute louvain_labels diffmap Manton_BM.h5ad test.html
    """

    def execute(self):
        kwargs = {
            "attr": self.args["--attribute"],
            "isreal": self.args["--is-real"],
            "isgene": self.args["--is-gene"],
            "log10": self.args["--log10"],
        }

        make_interactive_plots(
            self.args["<input_h5ad_file>"],
            self.args["<basis>"],
            self.args["<output_html_file>"],
            **kwargs
        )



import unittest

import pegasus as pg
import anndata
import numpy as np
from scipy.sparse import csr_matrix


class TestPreprocessing(unittest.TestCase):
    def test_log_norm(self):
        X = csr_matrix([[1, 11], [2, 20], [5, 6]])
        adata = anndata.AnnData(X)
        adata.var["robust"] = True
        pg.tools.log_norm(adata, 10)
        np.testing.assert_allclose(
            np.expm1(adata.X.toarray()).sum(axis=1), 10, rtol=1e-6, atol=0
        )


if __name__ == "__main__":
    unittest.main()



import gc
import json
import math

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import scipy.sparse
from pegasus.io import read_input
from pegasus.utils import decorators as pg_deco

obsm_whitelist = ['X_pca', 'X_rpca', 'X_tsne', 'X_fitsne', 'X_umap', 'X_fle', 'X_net_tsne', 'X_net_umap', 'X_net_fle']
obsm_whitelist_3d = ['X_diffmap_pca']


def create_schema(data):
    df = data.obs.iloc[[0, 1]].copy()
    obs_coords = []
    columns_to_add = []
    obsm = []
    for key in data.obsm.keys():
        ndim = None
        if key in obsm_whitelist:
            ndim = 2
        elif key in obsm_whitelist_3d:
            ndim = 3
        if ndim is not None:
            obs_coords.append(key)
            coordinate_columns = ['{}_{}'.format(key, i) for i in range(1, ndim + 1)]
            columns_to_add += coordinate_columns
            obsm.append({'name': key, 'dimensions': ndim})

    columns_to_add += data.var_names.to_list()
    empty_data = np.zeros((2, len(columns_to_add)), dtype='float32')
    empty_df = pd.DataFrame(index=df.index, columns=columns_to_add, data=empty_data)
    df = df.join(empty_df)
    table = pa.Table.from_pandas(df)
    schema = table.schema

    schema = schema.with_metadata(
        {b'pegasus': json.dumps(
            {'obsm': obsm, 'var': data.var.index.values.tolist(), 'obs': data.obs.columns.values.tolist()}).encode(
            'utf8')})
    return schema


def to_df(data, obs_subset=None):
    X = data.X if obs_subset is None else data.X[obs_subset]
    if scipy.sparse.issparse(X):
        X = X.toarray()

    if str(X.dtype) != 'float32':
        X = X.astype('float32')

    obs = data.obs if obs_subset is None else data.obs.iloc[obs_subset]
    df = pd.DataFrame(index=obs.index, data=X, columns=data.var_names)
    for key in data.obsm.keys():
        m = data.obsm[key] if obs_subset is None else data.obsm[key][obs_subset]

        if key in obsm_whitelist:
            df["{}_1".format(key)] = m[:, 0].astype('float32')
            df["{}_2".format(key)] = m[:, 1].astype('float32')
        elif key in obsm_whitelist_3d:
            df["{}_1".format(key)] = m[:, 0].astype('float32')
            df["{}_2".format(key)] = m[:, 1].astype('float32')
            df["{}_3".format(key)] = m[:, 2].astype('float32')
    df = df.join(data.obs)
    return df


def fix_categorical_dtypes(df):
    for column in df:
        if pd.api.types.is_categorical_dtype(df[column]) and len(df[column].dtype.categories) >= 10000:
            print('Converted {} to string'.format(column))
            df[column] = df[column].astype(str)


@pg_deco.TimeLogger()
def convert_to_parquet(data, output_name, nthreads, row_group_size):
    if not output_name.endswith(".pq") and not output_name.endswith(".parquet"):
        output_name = output_name + '.parquet'
    schema = create_schema(data)
    with pq.ParquetWriter(output_name, schema) as writer:
        for i in range(0, data.shape[0], row_group_size):
            end = i + row_group_size
            end = min(end, data.shape[0])
            df = to_df(data, slice(i, end))
            print('Converting {}-{}'.format(i, end))
            table = pa.Table.from_pandas(df, schema=schema, nthreads=nthreads)
            writer.write_table(table)
            gc.collect()
    print(output_name + " is written!")


def run_conversion(input_h5ad_file, output_name, nthreads, backed: bool = False, n_row_groups=None):
    data = read_input(input_h5ad_file, h5ad_mode='r' if backed else 'a')
    fix_categorical_dtypes(data.obs)
    if n_row_groups is None:
        n_row_groups = max(1, math.ceil(data.shape[0] / 100000))
    row_group_size = math.ceil(data.shape[0] / n_row_groups)
    convert_to_parquet(data, output_name, nthreads, row_group_size)



from pegasus.tools import run_conversion

from .Base import Base


class PARQUET(Base):
    """
Generate a PARQUET file for web-based visualization.

Usage:
  pegasus parquet [options] <input_h5ad_file> <output_name>
  pegasus parquet -h

Arguments:
  input_h5ad_file        Analyzed single cell data in h5ad format.
  output_name            Name prefix for the parquet file.

Options:
  -p <number>, --threads <number>           Number of threads used to generate the PARQUET file. [default: 1]
  -r <number>, --row_groups <number>        Number of parquet file row groups.
  -b, --backed                              Load h5ad file in backed mode instead of fully loading it into memory.
  -h, --help                                Print out help information.

Outputs:
  output_name.parquet        Generated PARQUET file that contains metadata and expression levels for every gene.

Examples:
  pegasus parquet -p 8 manton_bm.h5ad manton_bm
    """

    def execute(self):
        run_conversion(
            self.args["<input_h5ad_file>"],
            self.args["<output_name>"],
            nthreads=int(self.args["--threads"]),
            backed=self.args['--backed'],
            n_row_groups=int(self.args["--row_groups"]) if self.args.get("--row_groups", None) is not None else None
        )



import os
import unittest

import anndata
import h5py
import numpy as np
import pandas as pd
import pegasus as pg


class TestAggregate(unittest.TestCase):
    def tearDown(self):
        for file in ["aggregate_test.h5sc", "test1_dge.txt.gz", "test2_dge.txt.gz", "aggregate_test.csv"]:
            os.path.exists(file) and os.remove(file)

    def diff_m1_m2(self, result, m1, m2):
        self.assertEqual(result.shape[0], 2)
        self.assertEqual(result.shape[1], 5)
        # check that missing genes are 0 in aggregated result
        self.assertEqual(result['s1-c1'][:, ['d', 'e']].X.sum(), 0)
        self.assertEqual(result['s2-c1'][:, ['a', 'c']].X.sum(), 0)
        # check that genes maintain their original value
        self.assertEqual((result['s1-c1'][:, m1.var_names].X != m1.X).sum(), 0, "Values differ")
        self.assertEqual((result['s2-c1'][:, m2.var_names].X != m2.X).sum(), 0, "Values differ")

    def test_rename_reference(self):
        m1 = anndata.AnnData(X=np.random.random((1, 3)), var=pd.DataFrame(index=['a', 'b', 'c']),
            obs=pd.DataFrame(index=['c1']))
        pd.DataFrame(columns=m1.obs.index, index=m1.var.index, data=m1.X.T).to_csv('test1_dge.txt.gz', index_label='id',
            sep='\t')
        m2 = anndata.AnnData(X=np.random.random((1, 3)), var=pd.DataFrame(index=['b', 'd', 'e']),
            obs=pd.DataFrame(index=['c1']))
        pd.DataFrame(columns=m2.obs.index, index=m2.var.index, data=m2.X.T).to_csv('test2_dge.txt.gz', index_label='id',
            sep='\t')
        pd.DataFrame(
            data=dict(Sample=['s1', 's2'], Location=['test1_dge.txt.gz', 'test2_dge.txt.gz'],
                Reference=['b', 'c'], RenamedReference=['a', 'a'])).to_csv(
            'aggregate_test.csv', index=False)
        result = pg.aggregate_matrices(
            "aggregate_test.csv",
            what_to_return="AnnData"
        )
        self.diff_m1_m2(result, m1, m2)

    def test_union(self):
        m1 = anndata.AnnData(X=np.random.random((1, 3)), var=pd.DataFrame(index=['a', 'b', 'c']),
            obs=pd.DataFrame(index=['c1']))
        pd.DataFrame(columns=m1.obs.index, index=m1.var.index, data=m1.X.T).to_csv('test1_dge.txt.gz', index_label='id',
            sep='\t')
        m2 = anndata.AnnData(X=np.random.random((1, 3)), var=pd.DataFrame(index=['b', 'd', 'e']),
            obs=pd.DataFrame(index=['c1']))
        pd.DataFrame(columns=m2.obs.index, index=m2.var.index, data=m2.X.T).to_csv('test2_dge.txt.gz', index_label='id',
            sep='\t')
        pd.DataFrame(
            data=dict(Sample=['s1', 's2'], Location=['test1_dge.txt.gz', 'test2_dge.txt.gz'],
                Reference=['a', 'a'])).to_csv(
            'aggregate_test.csv', index=False)
        result = pg.aggregate_matrices(
            "aggregate_test.csv",
            what_to_return="AnnData"
        )
        self.diff_m1_m2(result, m1, m2)


    def test_aggregate_10x_matrices(self):
        m1 = pg.read_input(
            "tests/pegasus-test-data/input/heart_1k_v3/filtered_feature_bc_matrix.h5",
            genome="mm10",
        )
        m2 = pg.read_input(
            "tests/pegasus-test-data/input/heart_1k_v2/filtered_gene_bc_matrices_h5.h5",
            genome="mm10",
        )
        pg.aggregate_matrices(
            "tests/pegasus-test-data/input/aggregate_test.csv",
            what_to_return='aggregate_test',
        )

        result = pg.read_input("aggregate_test.h5sc", genome="mm10")
        self.assertEqual(
            m1.shape[0] + m2.shape[0], result.shape[0], "Cell dimension is incorrect"
        )
        self.assertEqual(m1.shape[1], result.shape[1], "Feature dimension is incorrect")

        m1_result = result[list(range(m1.shape[0])), :]
        m2_result = result[list(range(m1.shape[0], m1.shape[0] + m2.shape[0])), :]
        self.assertEqual((m1_result.X != m1.X).sum(), 0, "Values differ")
        self.assertEqual((m2_result.X != m2.X).sum(), 0, "Values differ")
        self.assertTrue(
            m1_result.obs.index.values[0].startswith("heart_1k_v3"), "Prefix not added"
        )
        self.assertTrue(
            m2_result.obs.index.values[0].startswith("heart_1k_v2"), "Prefix not added"
        )

    def test_multi_genome(self):
        pg.aggregate_matrices(
            "tests/pegasus-test-data/input/aggregate_multi_genome.csv",
            what_to_return='aggregate_test',
        )

        f = h5py.File("aggregate_test.h5sc", "r")
        self.assertIsNotNone(f["GRCh38"], "Genome not found")
        self.assertIsNotNone(f["mm10"], "Genome not found")
        with self.assertRaises(KeyError):
            f["mm9"]
        f.close()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, hstack, vstack
import tables

from typing import List
from pegasus.utils import decorators as pg_deco

import anndata


class Array2D:
    def __init__(
        self,
        barcode_metadata: "dict or pd.DataFrame" = {},
        feature_metadata: "dict or pd.DataFrame" = {},
        matrix: "csr_matrix" = csr_matrix((0, 0)),
        metadata: dict = {},
    ):
        self.barcode_metadata = (
            barcode_metadata
            if isinstance(barcode_metadata, pd.DataFrame)
            else pd.DataFrame(barcode_metadata)
        )
        self.feature_metadata = (
            feature_metadata
            if isinstance(feature_metadata, pd.DataFrame)
            else pd.DataFrame(feature_metadata)
        )
        self.matrix = matrix  # scipy csr matrix
        self.metadata = metadata  # other metadata, a dictionary

        if "barcodekey" in self.barcode_metadata:
            self.barcode_metadata.set_index("barcodekey", inplace=True)
        else:
            assert self.barcode_metadata.index.name == "barcodekey"
        if "featurekey" in self.feature_metadata:
            self.feature_metadata.set_index("featurekey", inplace=True)
        else:
            assert self.feature_metadata.index.name == "featurekey"

        if self.barcode_metadata.shape[0] != self.matrix.shape[0]:
            raise ValueError(
                "Wrong number of cells : matrix has {} cells, barcodes file has {}".format(
                    self.matrix.shape[0], self.barcode_metadata.shape[0]
                )
            )
        if self.feature_metadata.shape[0] != (
            self.matrix.shape[1] if len(self.matrix.shape) == 2 else 1
        ):
            raise ValueError(
                "Wrong number of features : matrix has {} features, features file has {}".format(
                    self.matrix.shape[1], self.feature_metadata.shape[0]
                )
            )

    def list_metadata_keys(self, type: str = "barcode") -> List[str]:
        """ Return available keys in metadata, type = barcode or feature or None
        """
        if type == "barcode":
            return [
                self.barcode_metadata.index.name
            ] + self.barcode_metadata.columns.tolist()
        elif type == "feature":
            return [
                self.feature_metadata.index.name
            ] + self.feature_metadata.columns.tolist()
        else:
            list(self.metadata)

    def get_metadata(self, name: str, type: str = "barcode") -> List[object]:
        """ Return values of column name in metadata type, type can be barcode, feature, or None
        """
        if type == "barcode":
            return (
                self.barcode_metadata.index
                if name == "barcodekey"
                else self.barcode_metadata[name].values
            )
        elif type == "feature":
            return (
                self.feature_metadata.index
                if name == "featurekey"
                else self.feature_metadata[name].values
            )
        else:
            return self.metadata[name]

    def trim(self, selected: List[bool]) -> None:
        """ Only keep barcodes in selected
        """
        self.matrix = self.matrix[selected, :]
        self.barcode_metadata = self.barcode_metadata[selected]

    def filter(self, ngene: int = None, select_singlets: bool = False) -> None:
        """ Filter out low quality barcodes, only keep barcodes satisfying ngene >= ngene and selecting singlets if select_singlets is True
        """
        if (ngene is None) and (not select_singlets):
            return None

        selected = np.ones(self.matrix.shape[0], dtype=bool)
        if ngene is not None:
            selected = selected & (self.matrix.getnnz(axis=1) >= ngene)
        if select_singlets:
            assert "demux_type" in self.barcode_metadata
            selected = (
                selected & (self.barcode_metadata["demux_type"] == "singlet").values
            )
            self.barcode_metadata.drop(columns="demux_type", inplace=True)

        self.trim(selected)

    def separate_channels(self, fn: str) -> None:
        """ Separate channel information from barcodekeys, only used for 10x v2, v3 h5 and mtx.
        """

        barcodes = self.barcode_metadata.index.values.astype(str)
        parts = np.array(np.char.rsplit(barcodes, sep="-", maxsplit=1).tolist())

        channels = None
        assert fn == "" or parts.shape[1] == 2
        if fn != "" and np.char.not_equal(parts[:, 1], "1").sum() > 0:
            # if we have multiple channels
            channels = [fn + "-" + x for x in parts[:, 1]]
            barcodes = np.array([x + "-" + y for x, y in zip(channels, parts[:, 0])])
        else:
            barcodes = parts[:, 0]

        self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")

        if channels is not None:
            self.barcode_metadata["Channel"] = channels

    def update_barcode_metadata_info(
        self, sample_name: str, row: "pd.Series", attributes: List[str]
    ) -> None:
        """ Update barcodekey, update channel and add attributes
        """
        nsample = self.barcode_metadata.shape[0]
        barcodes = [sample_name + "-" + x for x in self.barcode_metadata.index]
        self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")
        if "Channel" in self.barcode_metadata:
            self.barcode_metadata["Channel"] = [
                sample_name + "-" + x for x in self.barcode_metadata["Channel"]
            ]
        else:
            self.barcode_metadata["Channel"] = np.repeat(sample_name, nsample)
        if attributes is not None:
            for attr in attributes:
                self.barcode_metadata[attr] = np.repeat(row[attr], nsample)

    def write_to_hdf5(self, keyword: str, hd5_out: "File", is_h5sc: bool) -> None:
        """ Write Array2D content into hdf5 file
        """
        out_group = hd5_out.create_group("/", keyword)
        # write matrix
        hd5_out.create_carray(out_group, "data", obj=self.matrix.data)
        hd5_out.create_carray(out_group, "indices", obj=self.matrix.indices)
        hd5_out.create_carray(out_group, "indptr", obj=self.matrix.indptr)
        M, N = self.matrix.shape
        hd5_out.create_carray(
            out_group, "shape", obj=(N, M)
        )  # store as feature by barcode instead
        # write barcode_metadata
        outgb = hd5_out.create_group("/" + keyword, "_barcodes") if is_h5sc else out_group
        if not is_h5sc:
            self.barcode_metadata.index.name = "barcodes"
            self.feature_metadata.index.name = "genes"
            self.feature_metadata.rename(columns = {"featurename" : "gene_names"}, inplace = True)
        hd5_out.create_carray(
            outgb,
            self.barcode_metadata.index.name,
            obj=self.barcode_metadata.index.values.astype("S"),
        )  # encode into binary strings
        for col in self.barcode_metadata:
            kind = self.barcode_metadata[col].dtype.kind
            if kind == "U" or kind == "O":
                hd5_out.create_carray(
                    outgb, col, obj=self.barcode_metadata[col].values.astype("S")
                )  # encode into binary strings
            else:
                hd5_out.create_carray(outgb, col, obj=self.barcode_metadata[col].values)
        # write feature_metadata
        outgb = hd5_out.create_group("/" + keyword, "_features") if is_h5sc else out_group
        hd5_out.create_carray(
            outgb,
            self.feature_metadata.index.name,
            obj=self.feature_metadata.index.values.astype("S"),
        )  # encode into binary strings
        for col in self.feature_metadata:
            kind = self.feature_metadata[col].dtype.kind
            if kind == "U" or kind == "O":
                hd5_out.create_carray(
                    outgb, col, obj=self.feature_metadata[col].values.astype("S")
                )  # encode into binary strings
            else:
                hd5_out.create_carray(outgb, col, obj=self.feature_metadata[col].values)


def polish_featurename(
    feature_names: List[str], feature_keys: List[str], genomes: List[str]
) -> List[str]:
    """    Remove prefixing genome strings and deduplicate feature names
    """
    import re
    from collections import Counter

    prefix = re.compile("^(" + "|".join(genomes) + ")_+")
    if prefix.match(feature_names[0]):
        feature_names = [prefix.sub("", x) for x in feature_names]
        feature_keys = [prefix.sub("", x) for x in feature_keys]

    dup_ids = Counter()
    for i in range(len(feature_names)):
        idn = dup_ids[feature_names[i]]
        dup_ids[feature_names[i]] += 1
        if idn > 0:
            feature_names[i] = feature_names[i] + ".#~{}".format(idn + 1) # duplicate ID starts from 2. .#~ makes it unique.

    return feature_names, feature_keys


def get_fillna_dict(df: "pd.DataFrame") -> dict:
    """ Generate a fillna dict for columns in a df
    """
    fillna_dict = {}
    for column in df:
        if df[column].dtype.kind in {"O", "S"}:
            fillna_dict[column] = ""
        else:
            fillna_dict[column] = 0
    return fillna_dict


class MemData:
    def __init__(self):
        self.data = {}  # data is a dictionary mapping keyword to Array2D

    def listKeys(self) -> List[str]:
        return list(self.data)

    def getData(self, keyword: str) -> "Array2D":
        assert keyword in self.data
        return self.data[keyword]

    def addData(self, keyword: str, one_array: "Array2D") -> None:
        self.data[keyword] = one_array

    def update_barcode_metadata_info(
        self, sample_name: str, row: "pd.Series", attributes: List[str]
    ) -> None:
        """ Update barcodekey, update channel and add attributes for each array2d array
        """
        for array2d in self.data.values():
            array2d.update_barcode_metadata_info(sample_name, row, attributes)

    def addAggrData(self, data: "MemData") -> None:
        """ Add Aggr Data
        """
        for keyword, array2d in data.data.items():
            if keyword in self.data:
                self.data[keyword].append(array2d)
            else:
                self.data[keyword] = [array2d]

    @pg_deco.TimeLogger()
    def aggregate(self) -> None:
        """ Merge aggregated count matrices
        """
        import gc
        import warnings

        for keyword in self.data:
            array2d_list = self.data[keyword]
            self.data[keyword] = None

            if len(array2d_list) == 1:
                self.data[keyword] = array2d_list[0]
            else:
                barcode_metadata_dfs = [
                    array2d.barcode_metadata for array2d in array2d_list
                ]
                barcode_metadata = pd.concat(barcode_metadata_dfs, axis=0, sort=False)
                fillna_dict = get_fillna_dict(barcode_metadata)
                barcode_metadata.fillna(value=fillna_dict, inplace=True)

                feature_metadata = array2d_list[0].feature_metadata
                for other in array2d_list[1:]:
                    keys = ["featurekey"] + feature_metadata.columns.intersection(
                        other.feature_metadata.columns
                    ).values.tolist()
                    feature_metadata = feature_metadata.merge(
                        other.feature_metadata, on=keys, how="outer", sort=False
                    )  # If sort is True, feature keys will be changed even if all channels share the same feature keys.
                fillna_dict = get_fillna_dict(feature_metadata)
                feature_metadata.fillna(value=fillna_dict, inplace=True)

                matrix_list = []
                f2idx = pd.Series(
                    data=range(feature_metadata.shape[0]), index=feature_metadata.index
                )
                for array2d in array2d_list:
                    if (
                        feature_metadata.shape[0] > array2d.feature_metadata.shape[0]
                        or (
                            feature_metadata.index != array2d.feature_metadata.index
                        ).sum()
                        > 0
                    ):
                        mat = csr_matrix(
                            (array2d.matrix.shape[0], f2idx.size),
                            dtype=array2d.matrix.dtype,
                        )

                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            mat[
                                :, f2idx[array2d.feature_metadata.index].values
                            ] = array2d.matrix

                        array2d.matrix = mat
                        gc.collect()

                    matrix_list.append(array2d.matrix)

                newmat = vstack(matrix_list)
                matrix_list = array2d_list = None
                gc.collect()
                self.data[keyword] = Array2D(barcode_metadata, feature_metadata, newmat)

    def restrain_keywords(self, keywords: str) -> None:
        if keywords is None:
            return None

        keywords = set(keywords.split(","))
        available = set(self.data)

        invalid_set = keywords - available
        if len(invalid_set) > 0:
            raise ValueError(
                "Keywords {} do not exist.".format(",".join(list(invalid_set)))
            )

        remove_set = available - keywords
        for keyword in remove_set:
            self.data.pop(keyword)

    def convert_to_anndata(self, concat_matrices: bool =False, channel_attr: str = None, black_list: List[str] = []) -> "AnnData or List[AnnData]":
        """Convert an MemData object into SCANPY's AnnData object

        Parameters
        ----------

        concat_matrices : `bool`, optional (default: False)
            If concatenate multiple matrices. If so, return only one AnnData object, otherwise, might return a list of AnnData objects.

        channel_attr : `str`, optional (default: None)
            Use channel_attr to represent different samples. This will set a 'Channel' column field with channel_attr.

        black_list : `List[str]`, optional (default: [])
            Attributes in black list will be poped out.

        Returns
        -------

        `anndata` object or a dictionary of `anndata` objects
            An `anndata` object or a list of `anndata` objects containing the count matrices.

        Warning
        -------
        This procedure will convert all int matrix into float32 matrix in place!

        Examples
        --------
        >>> adata = MemData.convert_to_anndata()
        """

        Xs = []
        feature_dfs = []
        results = []

        genomes = self.listKeys()
        for genome in genomes:
            array2d = self.data[genome]

            if (
                array2d.matrix.dtype == np.int32
            ):  # caution, change matrix from int to float32 permanently!
                array2d.matrix.dtype = np.float32
                orig_data = array2d.matrix.data.view(np.int32)
                array2d.matrix.data[:] = orig_data

            obs_dict = array2d.barcode_metadata.to_dict(orient="list")
            obs_dict["obs_names"] = array2d.barcode_metadata.index.values
            if (channel_attr is not None) and (channel_attr in obs_dict):
                obs_dict["Channel"] = obs_dict[channel_attr]
            if "Channel" not in obs_dict:
                obs_dict["Channel"] = [""] * array2d.barcode_metadata.shape[0]
            for attr in black_list:
                if attr in obs_dict:
                    obs_dict.pop(attr)

            if not concat_matrices or len(genomes) == 1:
                var_dict = array2d.feature_metadata.to_dict(orient="list")
                feature_names, feature_keys = polish_featurename(
                    var_dict.pop("featurename"),
                    array2d.feature_metadata.index.values,
                    [genome],
                )
                var_dict["var_names"] = feature_names
                var_dict["gene_ids"] = feature_keys

                adata = anndata.AnnData(X=array2d.matrix, obs=obs_dict, var=var_dict)
                adata.uns["genome"] = genome
                results.append(adata)
            else:
                Xs.append(array2d.matrix)
                feature_dfs.append(array2d.feature_metadata)

        if len(results) == 0:
            fdf = pd.concat(feature_dfs, axis=0)
            fdf.fillna(value="N/A", inplace=True)
            var_dict = fdf.to_dict(orient="list")
            feature_names, feature_keys = polish_featurename(
                var_dict.pop("featurename"), fdf.index.values, genomes
            )
            var_dict["var_names"] = feature_names
            var_dict["gene_ids"] = feature_keys

            results = anndata.AnnData(
                X=hstack(Xs, format="csr"), obs=obs_dict, var=var_dict
            )
            results.uns["genome"] = ",".join(genomes)
        elif len(results) == 1:
            results = results[0]

        return results

    def write_h5_file(self, output_h5: str) -> None:
        """Write count matricies into a pegasus-format HDF5 file --- each matrix is stored in a group under root. For each matrix, the barcode metadata are stored under _barcodes subgroup and the feature metadata are stored under _features subgroup

        Parameters
        ----------

        output_h5 : `str`
            The output file name. If ends with .h5sc, output h5sc format; if ends with .h5, output 10x V2 h5 format.

        Returns
        -------

        None

        Examples
        --------
        >>> MemData.write_h5_file('example_10x.h5')
        """

        is_h5sc = output_h5.endswith(".h5sc")
        with tables.open_file(
            output_h5, mode="w", title=output_h5, filters=tables.Filters(complevel=1)
        ) as hd5_out:
            for keyword, array2d in self.data.items():
                array2d.write_to_hdf5(keyword, hd5_out, is_h5sc)




import pandas as pd
from matplotlib import pyplot as pl

from pegasus.io import read_input
from .plot_utils import transform_basis
from .plot_qc import plot_qc_violin
from . import plot_library, iplot_library




def make_static_plots(input_file, plot_type, output_file, dpi=500, **kwargs):
    adata = read_input(input_file, h5ad_mode="r")

    if plot_type == "qc_violin":
        if kwargs["attr"] is None:
            plot_qc_violin(
                adata,
                kwargs["qc_type"],
                output_file,
                xattr=kwargs["cluster"],
                xlabel=kwargs["cluster"],
                xtick_font=kwargs["qc_xtick_font"],
                xtick_rotation=kwargs["qc_xtick_rotation"],
                figsize=kwargs["subplot_size"],
                linewidth=kwargs["qc_line_width"],
            )
        else:
            plot_qc_violin(
                adata,
                kwargs["qc_type"],
                output_file,
                xattr=kwargs["cluster"],
                hue=kwargs["attr"],
                xlabel=kwargs["cluster"],
                xtick_font=kwargs["qc_xtick_font"],
                xtick_rotation=kwargs["qc_xtick_rotation"],
                split=True,
                figsize=kwargs["subplot_size"],
                linewidth=kwargs["qc_line_width"],
            )
    else:
        fig = getattr(plot_library, "plot_" + plot_type)(adata, **kwargs)
        fig.savefig(output_file, dpi=dpi)

    print(output_file + " is generated.")
    adata.file.close()


def make_interactive_plots(input_file, plot_type, output_file, **kwargs):
    adata = read_input(input_file, h5ad_mode="r")
    basis = transform_basis(plot_type)
    if plot_type == "diffmap" or plot_type == "diffmap_pca":
        df = pd.DataFrame(
            adata.obsm["X_{}".format(plot_type)][:, 0:3],
            index=adata.obs.index,
            columns=[basis + i for i in ["1", "2", "3"]],
        )
        if kwargs["isgene"]:
            coln = adata.var.index.get_loc(kwargs["attr"])
            df.insert(0, "Annotation", adata.X[:, coln].toarray().ravel())
        else:
            df.insert(0, "Annotation", adata.obs[kwargs["attr"]])
        if not kwargs["isreal"]:
            iplot_library.scatter3d(df, output_file)
        else:
            iplot_library.scatter3d_real(df, output_file, kwargs["log10"])
    else:
        df = pd.DataFrame(
            adata.obsm["X_{}".format(plot_type)],
            index=adata.obs.index,
            columns=[basis + i for i in ["1", "2"]],
        )
        if kwargs["isgene"]:
            coln = adata.var.index.get_loc(kwargs["attr"])
            df.insert(0, "Annotation", adata.X[:, coln].toarray().ravel())
        else:
            df.insert(0, "Annotation", adata.obs[kwargs["attr"]])
        if not kwargs["isreal"]:
            iplot_library.scatter(df, output_file)
        else:
            iplot_library.scatter_real(df, output_file, kwargs["log10"])
    print(output_file + " is generated.")
    adata.file.close()



import functools
import time
import gc
import logging

class TimeLogger:
    """
    Log time spent in a function.
    Can also log an additional custom mesage for debug / info / warning... purposes
    
    Example:
    -------

    @TimeLogger(custom_message="This a warning to be logged at the end", custom_message_type="warning")
    def print_something():
        print("Hello World")    

    """
    
    def __init__(self, custom_message:str=None, custom_message_type:str="debug", verbose=False):
        super(TimeLogger, self).__init__()
        self.custom_message = custom_message
        self.custom_message_type = custom_message_type
        self.verbose = verbose
        self.logger = logging.getLogger("pegasus")

    def __call__(self, function):
        if hasattr(function, "_pg_wrapped") :
            function = function._pg_wrapped
    
        @functools.wraps(function)
        def _do(*args, **kwargs):
            if self.verbose:
                self.logger.info(
                    "Entering: %s" % function.__code__.co_name
                )

            start = time.perf_counter()
            res = function(*args, **kwargs)
            end = time.perf_counter()
            
            self.logger.info(
                "Time spent on '{}' = {:.2f}s.".format(
                    function.__code__.co_name,
                    end - start
                )
            )

            if self.custom_message is not None:
                log_fct = getattr(self.logger, self.custom_message_type)
                log_fct(self.custom_message)
            
            return res

        _do._pg_wrapped = function
        return _do

class GCCollect:
    """
    Python's garbage collection is not that automatic. This decorator is
    especially useful for freeing memory after functions that call on modules
    that create a lot intermediary objects.

    Example:
    -------

    @GCCollect()
    def print_something():
        print("Hello World")

    """
    def __init__(self):
        super(GCCollect, self).__init__()
    
    def __call__(self, function):
        if hasattr(function, "_pg_wrapped") :
            function = function._pg_wrapped

        @functools.wraps(function)
        def _do(*args, **kwargs):
            res = function(*args, **kwargs)
            gc.collect()

            return res

        _do._pg_wrapped = function
        return _do



import os
import shutil
import unittest

import numpy as np
import pandas as pd

import pegasus as pg
from .test_util import assert_adata_equal


class TestRead(unittest.TestCase):
    def tearDown(self):
        for f in ["test.csv", "test.h5ad", "test_obsm_compound.h5ad"]:
            os.path.exists(f) and os.remove(f)
        os.path.exists('test_mtx') and shutil.rmtree('test_mtx')

    def test_mtx_v2(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/hgmm_1k_filtered_gene_bc_matrices/hg19/matrix.mtx"
        )
        self.assertEqual(adata.shape[0], 504)

    def test_mtx_v3(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/matrix.mtx.gz"
        )
        self.assertEqual(adata.shape[0], 1046)

    def test_mtx_v2_dir(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/hgmm_1k_filtered_gene_bc_matrices/hg19/"
        )
        self.assertEqual(adata.shape[0], 504)

    def test_write_mtx(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/heart_1k_v3/filtered_feature_bc_matrix.h5"
        )
        adata.var['test'] = 1.0
        adata.obs['test'] = 1.0
        output_dir = 'test_mtx/mm10'
        pg.write_output(adata, os.path.join(output_dir, 'matrix.mtx.gz'))
        adata2 = pg.read_input(output_dir)
        del adata2.obs['Channel']  # get channel from csv
        adata2.obs = adata2.obs.join(pd.read_csv(os.path.join(output_dir, 'obs.csv.gz'), index_col=0))
        adata2.var = adata2.var.join(pd.read_csv(os.path.join(output_dir, 'var.csv.gz'), index_col=0))
        del adata2.var['featuretype']
        assert_adata_equal(self, adata, adata2, obs_blacklist=['Channel'])

    def test_mtx_v3_dir(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/"
        )
        self.assertEqual(adata.shape[0], 1046)

    def test_csv(self):
        df = pd.DataFrame(index=["a", "b", "c"], data=dict(a=[1, 2, 3], b=[4, 5, 6]))
        df.to_csv("test.csv")
        adata = pg.read_input("test.csv", genome="test").T
        np.testing.assert_array_equal(df.values, adata.X.toarray())
        np.testing.assert_array_equal(df.index.values, adata.obs.index.values)
        np.testing.assert_array_equal(df.columns.values, adata.var.index.values)

        for chunk_size in [1, 2, 3, 4]:
            adata_chunks = pg.read_input(
                "test.csv", genome="test", chunk_size=chunk_size
            ).T
            assert_adata_equal(self, adata, adata_chunks)

    def test_read_write_h5ad(self):
        adata = pg.read_input(
            "tests/pegasus-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/"
        )
        pg.write_output(adata, "test.h5ad")
        adata2 = pg.read_input("test.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad(self):
        adata = pg.read_input("tests/pegasus-test-data/input/test_obsm_compound.h5ad")
        pg.write_output(adata, "test.h5ad")
        adata2 = pg.read_input("test.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad_backed(self):
        shutil.copy(
            "tests/pegasus-test-data/input/test_obsm_compound.h5ad",
            "test_obsm_compound.h5ad",
        )
        adata = pg.read_input("test_obsm_compound.h5ad", h5ad_mode="r+")
        pg.write_output(adata, "test_obsm_compound.h5ad")
        adata2 = pg.read_input("test_obsm_compound.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad_backed_whitelist(self):
        shutil.copy(
            "tests/pegasus-test-data/input/test_obsm_compound.h5ad",
            "test_obsm_compound.h5ad",
        )
        adata = pg.read_input("test_obsm_compound.h5ad", h5ad_mode="r+")
        pg.write_output(adata, "test_obsm_compound.h5ad", whitelist=["obs"])
        adata2 = pg.read_input("test_obsm_compound.h5ad")
        assert_adata_equal(self, adata, adata2)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
from typing import List
from anndata import AnnData

import logging
logger = logging.getLogger("pegasus")

from pegasus.io import read_input


def search_genes(
    data: AnnData,
    gene_list: List[str],
    rec_key: str = "de_res",
    measure: str = "percentage",
) -> pd.DataFrame:
    """Extract and display gene expressions for each cluster from an `anndata` object.

    This function helps to see marker expressions in clusters via the interactive python environment.

    Parameters
    ----------

    data: ``anndata.AnnData``
        Annotated data matrix containing the expression matrix and differential expression results.

    gene_list: ``List[str]``
        A list of gene symbols.

    rec_key: ``str``, optional, default: ``"de_res"``
        Keyword of DE analysis result stored in ``data.varm``.

    measure : ``str``, optional, default: ``"percentage"``
        Can be either ``"percentage"`` or ``"mean_logExpr"``:
            * ``percentage`` shows the percentage of cells expressed the genes;
            * ``mean_logExpr`` shows the mean log expression.

    Returns
    -------
    ``pandas.DataFrame``
        A data frame containing marker expressions in each cluster.

    Examples
    --------
    >>> results = pg.search_genes(adata, ['CD3E', 'CD4', 'CD8'])
    """

    columns = [x for x in data.varm[rec_key].dtype.names if x.startswith(measure + ":")]
    df = pd.DataFrame(data=data.varm[rec_key][columns], index=data.var_names)
    return df.reindex(index=gene_list)


def search_de_genes(
    data: AnnData,
    gene_list: List[str],
    rec_key: str = "de_res",
    de_test: str = "fisher",
    de_alpha: float = 0.05,
    thre: float = 1.5,
) -> pd.DataFrame:
    """Extract and display differential expression analysis results of markers for each cluster.

    This function helps to see if markers are up or down regulated in each cluster via the interactive python environment: 
        * ``++`` indicates up-regulated and fold change >= threshold;
        * ``+`` indicates up-regulated but fold change < threshold;
        * ``--`` indicates down-regulated and fold change <= 1 / threshold; 
        * ``-`` indicates down-regulated but fold change > 1 / threshold;
        * ``?`` indicates not differentially expressed.

    Parameters
    ----------
    data: ``anndata.Anndata``
        Annotated data matrix containing the expression matrix and differential expression results.

    gene_list: ``List[str]``
        A list of gene symbols.

    rec_key: ``str``, optional, default: ``"de_res"``
        Keyword of DE analysis result stored in ``data.varm``.

    de_test : ``str``, optional, default: ``"fisher"``
        Differential expression test to look at, could be either ``t``, ``fisher`` or ``mwu``.

    de_alpha : ``float``, optional, default: ``0.05``
        False discovery rate.

    thre : ``float``, optional, default: ``1.5``
        Fold change threshold to determine if the marker is a strong DE (``++`` or ``--``) or weak DE (``+`` or ``-``).

    Returns
    -------
    ``pandas.DataFrame``
        A data frame containing marker differential expression results for each cluster.

    Examples
    --------
    >>> df = pegasus.misc.search_de_genes(adata, ['CD3E', 'CD4', 'CD8'], thre = 2.0)
    """

    columns = [
        x for x in data.varm[rec_key].dtype.names if x.startswith(de_test + "_qval:")
    ]
    df_de = pd.DataFrame(data.varm[rec_key][columns], index=data.var_names)
    df_de = df_de.reindex(index=gene_list)

    columns = [
        x
        for x in data.varm[rec_key].dtype.names
        if (
            x.startswith("percentage_fold_change:")
            if de_test == "fisher"
            else x.startswith("log_fold_change:")
        )
    ]
    df_fc = pd.DataFrame(data.varm[rec_key][columns], index=data.var_names)
    df_fc = df_fc.reindex(index=gene_list)
    if de_test != "fisher":
        df_fc = np.exp(df_fc)

    results = np.zeros((len(gene_list), len(columns)), dtype=np.dtype("U4"))
    results[:] = "?"
    results[np.isnan(df_de)] = "NaN"
    results[(df_de <= de_alpha).values & (df_fc > 1.0).values] = "+"
    results[(df_de <= de_alpha).values & (df_fc >= thre).values] = "++"
    results[(df_de <= de_alpha).values & (df_fc < 1.0).values] = "-"
    results[(df_de <= de_alpha).values & (df_fc <= 1.0 / thre).values] = "--"

    clusts = [x.rpartition(":")[2] for x in columns]
    df = pd.DataFrame(data=results, index=gene_list, columns=clusts)
    return df


def show_attributes(
    input_file: str,
    show_attributes: bool,
    show_gene_attributes: bool,
    show_values_for_attributes: str,
) -> None:
    """ Show data attributes. For command line use.
    """

    data = read_input(input_file, h5ad_mode="r")
    if show_attributes:
        print(
            "Available sample attributes in input dataset: {0}".format(
                ", ".join(data.obs.columns.values)
            )
        )
    if show_gene_attributes:
        print(
            "Available gene attributes in input dataset: {0}".format(
                ", ".join(data.var.columns.values)
            )
        )
    if not show_values_for_attributes is None:
        for attr in show_values_for_attributes.split(","):
            print(
                "Available values for attribute {0}: {1}.".format(
                    attr, ", ".join(np.unique(data.obs[attr]))
                )
            )


def perform_oneway_anova(
    data: AnnData,
    glist: List[str],
    restriction_vec: List[str],
    group_str: str,
    fdr_alpha: float = 0.05,
    res_key: str = None,
) -> pd.DataFrame:
    """Perform one way ANOVA on a subset of cells (restricted by restriction_vec) grouped by group_str and control FDR at fdr_alpha.
    Parameters
    ----------

    data : `anndata` object
        An `anndata` object containing the expression matrix.
    glist : `list[str]`
        A list of gene symbols.
    restriction_vec : `list[str]`
        A vector of restrictions for selecting cells. Each restriction takes the format of attr:value,value,value
    group_str : `str`
        How to group selected cells for ANOVA analysis. If group_str is for pseudotime, it has two formats. 1) 'pseudotime:time:n', which divides cells by equal pseudotime invertal; 2) 'pseudotime:size:n' divides cells by equal number of cells.
    fdr_alpha : `float`, optional (default: 0.05)
        False discovery rate.
    res_key : `str`, optional (default: None)
        Store results into data using res_key, the grouping information is stored in obs and the results is stored in uns.

    Returns
    -------
    `pandas.DataFrame`
        Results for genes that pass FDR control.

    Examples
    --------
    >>> results = misc.perform_oneway_anova(data, ['CD3E', 'CD4', 'CD8'], [], 'pseudotime:size:10')
    """

    from scipy.stats import f_oneway
    from statsmodels.stats.multitest import fdrcorrection as fdr

    selected = np.ones(data.shape[0], dtype=bool)
    for rest_str in restriction_vec:
        attr, value_str = rest_str.split(":")
        values = value_str.split(",")
        selected = selected & np.isin(data.obs[attr], values)

    gene_list = np.array(glist)
    gene_list = gene_list[np.isin(gene_list, data.var_names)]
    ngene = gene_list.size

    newdat = data[selected, :][:, gene_list].copy()
    newdat.X = newdat.X.toarray()

    group_values = group_str.split(":")
    group_names = []
    col_names = []

    ngr = 0
    group_idx = None

    if group_values[0] == "pseudotime":
        assert len(group_values) == 3
        div_by = group_values[1]
        ngr = int(group_values[2])

        group_idx = np.zeros((ngr, newdat.shape[0]), dtype=bool)
        pseudotimes = newdat.obs["pseudotime"].values

        min_t = pseudotimes.min()
        max_t = pseudotimes.max()

        if div_by == "time":
            interval = (max_t - min_t) / ngr
            left = min_t - 1e-5
            for i in range(ngr):
                right = min_t + interval * (i + 1)
                name = "({:.2f}, {:.2f}]".format(left if left >= 0 else 0.0, right)
                group_names.append(name)
                group_idx[i] = (pseudotimes > left) & (pseudotimes <= right)
                left = right
        else:
            assert div_by == "size"
            ords = np.argsort(pseudotimes)
            quotient = ords.size // ngr
            residule = ords.size % ngr

            fr = 0
            for i in range(ngr):
                to = fr + quotient + (i < residule)
                name = "[{:.2f}, {:.2f}]".format(
                    pseudotimes[ords[fr]], pseudotimes[ords[to - 1]]
                )
                group_names.append(name)
                group_idx[i][ords[fr:to]] = True
                fr = to

    else:
        assert len(group_values) == 2
        group_attr = group_values[0]
        tmp_str = group_values[1]
        groups_str = tmp_str.split(";")

        ngr = len(groups_str)
        group_idx = np.zeros((ngr, newdat.shape[0]), dtype=bool)

        for i, gstr in enumerate(groups_str):
            name, values = gstr.split("~")
            group_names.append(name)
            group_idx[i] = np.isin(newdat.obs[group_attr], values.split(","))

    for i in range(ngr):
        print("Group {} has {} cells.".format(group_names[i], group_idx[i].sum()))

    np.warnings.filterwarnings("ignore")
    stats = np.zeros((ngene, 3 + ngr * 2))
    for i in range(ngene):
        arr_list = []
        for j in range(ngr):
            arr = newdat.X[group_idx[j], i]
            stats[i, 3 + j * 2] = arr.mean()
            stats[i, 3 + j * 2 + 1] = (arr > 0).sum() * 100.0 / arr.size
            arr_list.append(arr)
        stats[i, 0], stats[i, 1] = f_oneway(*arr_list)
        if np.isnan(stats[i, 0]):
            stats[i, 0] = 0.0
            stats[i, 1] = 1.0
    passed, stats[:, 2] = fdr(stats[:, 1])

    cols = ["fstat", "pval", "qval"]
    for i in range(ngr):
        cols.extend([group_names[i] + "_mean", group_names[i] + "_percent"])
    raw_results = pd.DataFrame(stats, columns=cols, index=gene_list)

    results = raw_results[raw_results["qval"] <= fdr_alpha]
    results = results.sort_values("qval")

    if res_key is not None:
        data.uns[res_key] = raw_results
        data.obs[res_key] = "background"
        for i in range(ngr):
            idx = np.zeros(data.shape[0], dtype=bool)
            idx[selected] = group_idx[i]
            data.obs.loc[idx, res_key] = group_names[i]

    return results



def remap_singlets(data: AnnData, remap_string: str) -> None:
    """ Remap demuxEM identified singlets. For example cell1_tag1, cell1_tag2 can be mapped all to cell1.
    """
    if 'assignment' not in data.obs:
        raise ValueError("No assignment field detected!")
    data.obs['assignment.orig'] = data.obs['assignment']
    remap = {}
    tokens = remap_string.split(';')
    for token in tokens:
        new_key, old_str = token.split(':')
        old_keys = old_str.split(',')
        for key in old_keys:
            remap[key] = new_key
    data.obs['assignment'] = pd.Categorical(data.obs['assignment'].apply(lambda x: remap[x] if x in remap else x))
    logger.info("Singlets are remapped.")


def subset_singlets(data: AnnData, subset_string: str) -> None:
    """ Select subset of singlets.
    """
    if 'assignment' not in data.obs:
        raise ValueError("No assignment field detected!")
    subset = np.array(subset_string.split(','))
    idx = np.isin(data.obs['assignment'], subset)
    n_before = data.shape[0]
    data._inplace_subset_obs(idx)
    logger.info("Singlets are subsetted. {0} singlets are selected from {1} cells.".format(idx.sum(), n_before))



# This code is used to write h5ad 'r+' modified attributes back to disk. The code is adapted from AnnData codebase

import numpy as np
import pandas as pd
from pandas.api.types import is_categorical_dtype, is_string_dtype
import h5py
from collections.abc import Mapping
from natsort import natsorted
from scipy.sparse import issparse
import anndata
from typing import List, Tuple



# string dtype
sdt = h5py.string_dtype(encoding = 'utf-8')



def _df_to_records_fixed_width(df: pd.DataFrame) -> Tuple[np.recarray, dict]:
    names = ['index']
    arrays = [df.index.values.astype(sdt)]
    uns_cat = {}
    for col in df.columns:
        names.append(col)
        values = df[col].values
        if is_categorical_dtype(values):
            uns_cat[col + '_categories'] = values.categories
            arrays.append(values.codes)
        elif is_string_dtype(values):
            keywords = np.unique(values)
            if keywords.size < values.size:
                values = pd.Categorical(values, categories = natsorted(keywords))
                uns_cat[col + '_categories'] = values.categories
                arrays.append(values.codes)
            else:
                arrays.append(values.astype(sdt))
        else:
            arrays.append(values)
    formats = [values.dtype for values in arrays]
    return np.rec.fromarrays(arrays, dtype = {'names' : names, 'formats' : formats}), uns_cat


def _to_dict_fixed_width_arrays(data: anndata) -> dict:
    obs_rec, uns_obs = _df_to_records_fixed_width(data.obs)
    var_rec, uns_var = _df_to_records_fixed_width(data.var)
    res = {
        'X': data.X,
        'obs': obs_rec,
        'var': var_rec,
        'obsm': data.obsm,
        'varm': data.varm,
        'uns': {**data.uns, **uns_obs, **uns_var}
    }
    return res



def _parse_whitelist(whitelist: List[str]) -> dict:
    parse_results = {}
    for value in whitelist:
        tokens = value.split("/")
        curr_dict = parse_results
        for i in range(len(tokens) - 1):
            if tokens[i] not in curr_dict:
                curr_dict[tokens[i]] = dict()
            curr_dict = curr_dict[tokens[i]]
            if curr_dict is None:
                break
        if curr_dict is not None:
            curr_dict[tokens[-1]] = None
    return parse_results


def _update_backed_h5ad(group: "hdf5 group", dat: dict, whitelist: dict):
    for key, value in dat.items():
        if not isinstance(key, str):
            logging.warning(
                "Dictionary key {} is transformed to str upon writing to h5,"
                "using string keys is recommended".format(key)
            )
            key = str(key)

        if whitelist is None or key in whitelist:
            if isinstance(value, Mapping):
                subgroup = (
                    group[key] if key in group.keys() else group.create_group(key)
                )
                assert isinstance(subgroup, h5py.Group)
                _update_backed_h5ad(
                    subgroup, value, whitelist[key] if whitelist is not None else None
                )
            else:
                if key in group.keys():
                    del group[key]
                if issparse(value):
                    sparse_mat = group.create_group(key)
                    sparse_mat.attrs["h5sparse_format"] = value.format
                    sparse_mat.attrs["h5sparse_shape"] = np.array(value.shape)
                    sparse_mat.create_dataset(
                        "data", data=value.data, compression="gzip"
                    )
                    sparse_mat.create_dataset(
                        "indices", data=value.indices, compression="gzip"
                    )
                    sparse_mat.create_dataset(
                        "indptr", data=value.indptr, compression="gzip"
                    )
                else:
                    value = np.array(value) if np.ndim(value) > 0 else np.array([value])
                    if value.dtype.kind in {"U", "O"}:
                        value = value.astype(sdt)
                    if value.dtype.names is not None:
                        new_dtype = value.dtype.descr
                        convert_type = False
                        for i in range(len(value.dtype)):
                            if value.dtype[i].kind in {"U", "O"}:
                                new_dtype[i] = (new_dtype[i][0], sdt)
                                convert_type = True
                        if convert_type:
                            value = value.astype(new_dtype)
                    group.create_dataset(key, data=value, compression="gzip")



import numpy as np
import anndata

from typing import List
import logging

logger = logging.getLogger("pegasus")


def parse_subset_selections(subset_selections):
    subsets_dict = {}
    for subset_str in subset_selections:
        attr, value_str = subset_str.split(":")
        if attr in subsets_dict:
            subsets_dict[attr].extend(value_str.split(","))
        else:
            subsets_dict[attr] = value_str.split(",")
    return subsets_dict


def get_anndata_for_subclustering(data: "AnnData", subset_selections: List[str]):
    obs_index = np.full(data.shape[0], True)
    subsets_dict = parse_subset_selections(subset_selections)
    for key, value in subsets_dict.items():
        logger.info("{} in {}".format(str(key), str(value)))
        obs_index = obs_index & np.isin(data.obs[key], value)
    data = data[obs_index, :]

    obs_dict = {"obs_names": data.obs_names.values}
    for attr in data.obs.columns:
        if attr != "pseudotime":
            if attr.find("_labels") < 0:
                obs_dict[attr] = data.obs[attr].values
            else:
                obs_dict["parent_" + attr] = data.obs[attr].values

    var_dict = {
        "var_names": data.var_names.values,
        "gene_ids": data.var["gene_ids"].values,
        "robust": data.var["robust"].values,
    }

    newdata = anndata.AnnData(X=data.X, obs=obs_dict, var=var_dict)
    newdata.var["n_cells"] = newdata.X.getnnz(axis=0)
    newdata.var["robust"] = (
        newdata.var["robust"].values & (newdata.var["n_cells"] > 0).values
    )
    newdata.var["highly_variable_features"] = newdata.var[
        "robust"
    ]  # default all robust genes are "highly" variable

    if "Channels" in data.uns:
        newdata.uns["Channels"] = data.uns["Channels"]
    if "Groups" in data.uns:
        newdata.uns["Groups"] = data.uns["Groups"]
    if "plus" in data.varm.keys():
        newdata.varm["plus"] = data.varm["plus"]
    if "muls" in data.varm.keys():
        newdata.varm["muls"] = data.varm["muls"]

    logger.info("{0} cells are selected.".format(newdata.shape[0]))

    return newdata



import time
import numpy as np
import pandas as pd
from anndata import AnnData
from joblib import effective_n_jobs
from natsort import natsorted

import ctypes
import ctypes.util

from sklearn.cluster import KMeans
from typing import List

from pegasus.tools import construct_graph

import logging
logger = logging.getLogger("pegasus")
from pegasus.utils import decorators as pg_deco


def louvain(
    data: AnnData,
    rep: str = "pca",
    resolution: int = 1.3,
    random_state: int = 0,
    class_label: str = "louvain_labels",
) -> None:
    """Cluster the cells using Louvain algorithm.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation used for clustering. Keyword ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    resolution: ``int``, optional, default: ``1.3``
        Resolution factor. Higher resolution tends to find more clusters with smaller sizes.

    random_state: ``int``, optional, default: ``0``
        Random seed for reproducing results.

    class_label: ``str``, optional, default: ``"louvain_labels"``
        Key name for storing cluster labels in ``data.obs``.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[class_label]``: Cluster labels of cells as categorical data.

    Examples
    --------
    >>> pg.louvain(adata)
    """

    try:
        import louvain as louvain_module
    except ImportError:
        print("Need louvain! Try 'pip install louvain-github'.")

    start = time.perf_counter()

    rep_key = "W_" + rep
    if rep_key not in data.uns:
        raise ValueError("Cannot find affinity matrix. Please run neighbors first!")
    W = data.uns[rep_key]

    G = construct_graph(W)
    partition_type = louvain_module.RBConfigurationVertexPartition
    partition = partition_type(G, resolution_parameter=resolution, weights="weight")
    optimiser = louvain_module.Optimiser()
    optimiser.set_rng_seed(random_state)
    diff = optimiser.optimise_partition(partition)

    labels = np.array([str(x + 1) for x in partition.membership])
    categories = natsorted(np.unique(labels))
    data.obs[class_label] = pd.Categorical(values=labels, categories=categories)

    end = time.perf_counter()
    logger.info("Louvain clustering is done. Time spent = {:.2f}s.".format(end - start))

def leiden(
    data: AnnData,
    rep: str = "pca",
    resolution: int = 1.3,
    n_iter: int = -1,
    random_state: int = 0,
    class_label: str = "leiden_labels",
) -> None:
    """Cluster the data using Leiden algorithm.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation used for clustering. Keyword ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    resolution: ``int``, optional, default: ``1.3``
        Resolution factor. Higher resolution tends to find more clusters.

    n_iter: ``int``, optional, default: ``-1``
        Number of iterations that Leiden algorithm runs. If ``-1``, run the algorithm until reaching its optimal clustering.

    random_state: ``int``, optional, default: ``0``
        Random seed for reproducing results.

    class_label: ``str``, optional, default: ``"leiden_labels"``
        Key name for storing cluster labels in ``data.obs``.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[class_label]``: Cluster labels of cells as categorical data.

    Examples
    --------
    >>> pg.leiden(adata)
    """

    try:
        import leidenalg
    except ImportError:
        print("Need leidenalg! Try 'pip install leidenalg'.")

    start = time.perf_counter()

    rep_key = "W_" + rep
    if rep_key not in data.uns:
        raise ValueError("Cannot find affinity matrix. Please run neighbors first!")
    W = data.uns[rep_key]

    G = construct_graph(W)
    partition_type = leidenalg.RBConfigurationVertexPartition
    partition = leidenalg.find_partition(
        G,
        partition_type,
        seed=random_state,
        weights="weight",
        resolution_parameter=resolution,
        n_iterations=n_iter,
    )

    labels = np.array([str(x + 1) for x in partition.membership])
    categories = natsorted(np.unique(labels))
    data.obs[class_label] = pd.Categorical(values=labels, categories=categories)

    end = time.perf_counter()
    logger.info("Leiden clustering is done. Time spent = {:.2f}s.".format(end - start))


@pg_deco.TimeLogger()
def partition_cells_by_kmeans(data: AnnData, rep: str, n_jobs: int, n_clusters: int, n_clusters2: int, n_init: int, random_state: int) -> List[int]:
    n_jobs = effective_n_jobs(n_jobs)

    rep_key = "X_" + rep
    X = data.obsm[rep_key].astype("float64")

    km = KMeans(n_clusters=n_clusters, n_jobs=n_jobs, n_init = n_init, random_state=random_state)
    km.fit(X)
    coarse = km.labels_.copy()

    km.set_params(n_init = 1)
    labels = coarse.copy()
    base_sum = 0
    for i in range(n_clusters):
        idx = coarse == i
        nc = min(n_clusters2, idx.sum())
        km.set_params(n_clusters=nc)
        km.fit(X[idx,:])
        labels[idx] = base_sum + km.labels_
        base_sum += nc

    return labels


def spectral_louvain(
    data: AnnData,
    rep: str = "pca",
    resolution: float = 1.3,
    rep_kmeans: str = "diffmap",
    n_clusters: int = 30,
    n_clusters2: int = 50,
    n_init: int = 10,
    n_jobs: int = -1,
    random_state: int = 0,
    class_label: str = "spectral_louvain_labels",
) -> None:
    """ Cluster the data using Spectral Louvain algorithm.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation used for clustering. Keyword ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    resolution: ``int``, optional, default: ``1.3``
        Resolution factor. Higher resolution tends to find more clusters with smaller sizes.

    rep_kmeans: ``str``, optional, default: ``"diffmap"``
        The embedding representation on which the KMeans runs. Keyword must exist in ``data.obsm``. By default, use Diffusion Map coordinates. If diffmap is not calculated, use PCA coordinates instead.

    n_clusters: ``int``, optional, default: ``30``
        The number of first level clusters.

    n_clusters2: ``int``, optional, default: ``50``
        The number of second level clusters.

    n_init: ``int``, optional, default: ``10``
        Number of kmeans tries for the first level clustering. Default is set to be the same as scikit-learn Kmeans function.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all available threads.

    random_state: ``int``, optional, default: ``0``
        Random seed for reproducing results.

    class_label: ``str``, optional, default: ``"spectral_louvain_labels"``
        Key name for storing cluster labels in ``data.obs``.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[class_label]``: Cluster labels for cells as categorical data.

    Examples
    --------
    >>> pg.spectral_louvain(adata)
    """

    try:
        import louvain as louvain_module
    except ImportError:
        print("Need louvain! Try 'pip install louvain-github'.")

    start = time.perf_counter()

    if "X_" + rep_kmeans not in data.obsm.keys():
        logger.warning("{} is not calculated, switch to pca instead.".format(rep_kmeans))
        rep_kmeans = "pca"
        if "X_" + rep_kmeans not in data.obsm.keys():
            raise ValueError("Please run {} first!".format(rep_kmeans))
    if "W_" + rep not in data.uns:
        raise ValueError("Cannot find affinity matrix. Please run neighbors first!")

    labels = partition_cells_by_kmeans(
        data, rep_kmeans, n_jobs, n_clusters, n_clusters2, n_init, random_state,
    )

    W = data.uns["W_" + rep]

    G = construct_graph(W)
    partition_type = louvain_module.RBConfigurationVertexPartition
    partition = partition_type(
        G, resolution_parameter=resolution, weights="weight", initial_membership=labels
    )
    partition_agg = partition.aggregate_partition()

    optimiser = louvain_module.Optimiser()
    optimiser.set_rng_seed(random_state)
    diff = optimiser.optimise_partition(partition_agg)
    partition.from_coarse_partition(partition_agg)

    labels = np.array([str(x + 1) for x in partition.membership])
    categories = natsorted(np.unique(labels))
    data.obs[class_label] = pd.Categorical(values=labels, categories=categories)

    end = time.perf_counter()
    logger.info(
        "Spectral Louvain clustering is done. Time spent = {:.2f}s.".format(end - start)
    )


def spectral_leiden(
    data: AnnData,
    rep: str = "pca",
    resolution: float = 1.3,
    rep_kmeans: str = "diffmap",
    n_clusters: int = 30,
    n_clusters2: int = 50,
    n_init: int = 10,
    n_jobs: int = -1,
    random_state: int = 0,
    class_label: str = "spectral_leiden_labels",
) -> None:
    """Cluster the data using Spectral Leiden algorithm.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation used for clustering. Keyword ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    resolution: ``int``, optional, default: ``1.3``
        Resolution factor. Higher resolution tends to find more clusters.

    rep_kmeans: ``str``, optional, default: ``"diffmap"``
        The embedding representation on which the KMeans runs. Keyword must exist in ``data.obsm``. By default, use Diffusion Map coordinates. If diffmap is not calculated, use PCA coordinates instead.

    n_clusters: ``int``, optional, default: ``30``
        The number of first level clusters.

    n_clusters2: ``int``, optional, default: ``50``
        The number of second level clusters.

    n_init: ``int``, optional, default: ``10``
        Number of kmeans tries for the first level clustering. Default is set to be the same as scikit-learn Kmeans function.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all available threads.

    random_state: ``int``, optional, default: ``0``
        Random seed for reproducing results.

    class_label: ``str``, optional, default: ``"spectral_leiden_labels"``
        Key name for storing cluster labels in ``data.obs``.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[class_label]``: Cluster labels for cells as categorical data.

    Examples
    --------
    >>> pg.spectral_leiden(adata)
    """

    try:
        import leidenalg
    except ImportError:
        print("Need leidenalg! Try 'pip install leidenalg'.")

    start = time.perf_counter()

    if "X_" + rep_kmeans not in data.obsm.keys():
        logger.warning("{} is not calculated, switch to pca instead.".format(rep_kmeans))
        rep_kmeans = "pca"
        if "X_" + rep_kmeans not in data.obsm.keys():
            raise ValueError("Please run {} first!".format(rep_kmeans))
    if "W_" + rep not in data.uns:
        raise ValueError("Cannot find affinity matrix. Please run neighbors first!")

    labels = partition_cells_by_kmeans(
        data, rep_kmeans, n_jobs, n_clusters, n_clusters2, n_init, random_state,
    )

    W = data.uns["W_" + rep]

    G = construct_graph(W)
    partition_type = leidenalg.RBConfigurationVertexPartition
    partition = partition_type(
        G, resolution_parameter=resolution, weights="weight", initial_membership=labels
    )
    partition_agg = partition.aggregate_partition()

    optimiser = leidenalg.Optimiser()
    optimiser.set_rng_seed(random_state)
    diff = optimiser.optimise_partition(partition_agg, -1)
    partition.from_coarse_partition(partition_agg)

    labels = np.array([str(x + 1) for x in partition.membership])
    categories = natsorted(np.unique(labels))
    data.obs[class_label] = pd.Categorical(values=labels, categories=categories)

    end = time.perf_counter()
    logger.info(
        "Spectral Leiden clustering is done. Time spent = {:.2f}s.".format(end - start)
    )


def cluster(
    data: AnnData,
    algo: str = "louvain",
    rep: str = "pca",
    resolution: int = 1.3,
    random_state: int = 0,
    class_label: str = None,
    n_iter: int = -1,
    rep_kmeans: str = "diffmap",
    n_clusters: int = 30,
    n_clusters2: int = 50,
    n_init: int = 10,
    n_jobs: int = -1,
) -> None:
    """Cluster the data using the chosen algorithm. Candidates are louvain, leiden, spectral_louvain and spectral_leiden. If data have < 1000 cells and there are clusters with sizes of 1, resolution is automatically reduced until no cluster of size 1 appears.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    algo: ``str``, optional, default: ``"louvain"``
        Which clustering algorithm to use. Choices are louvain, leiden, spectral_louvain, spectral_leiden

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation used for clustering. Keyword ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    resolution: ``int``, optional, default: ``1.3``
        Resolution factor. Higher resolution tends to find more clusters.

    random_state: ``int``, optional, default: ``0``
        Random seed for reproducing results.

    class_label: ``str``, optional, default: None
        Key name for storing cluster labels in ``data.obs``. If None, use 'algo_labels'.

    n_iter: ``int``, optional, default: ``-1``
        Number of iterations that Leiden algorithm runs. If ``-1``, run the algorithm until reaching its optimal clustering.

    rep_kmeans: ``str``, optional, default: ``"diffmap"``
        The embedding representation on which the KMeans runs. Keyword must exist in ``data.obsm``. By default, use Diffusion Map coordinates. If diffmap is not calculated, use PCA coordinates instead.

    n_clusters: ``int``, optional, default: ``30``
        The number of first level clusters.

    n_clusters2: ``int``, optional, default: ``50``
        The number of second level clusters.

    n_init: ``int``, optional, default: ``10``
        Number of kmeans tries for the first level clustering. Default is set to be the same as scikit-learn Kmeans function.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all available threads.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[class_label]``: Cluster labels of cells as categorical data.

    Examples
    --------
    >>> pg.cluster(adata, algo = 'leiden')
    """

    if algo not in {'louvain', 'leiden', 'spectral_louvain', 'spectral_leiden'}:
        raise ValueError("Unknown clustering algorithm {}.".format(algo))

    if class_label is None:
        class_label = algo + '_labels'

    kwargs = {'data' : data, 'rep' : rep, 'resolution' : resolution, 'random_state' : random_state, 'class_label': class_label}
    if algo == 'leiden':
        kwargs['n_iter'] = n_iter
    if algo in ['spectral_louvain', 'spectral_leiden']:
        kwargs.update({'rep_kmeans': rep_kmeans, 'n_clusters': n_clusters, 'n_clusters2': n_clusters2, 'n_init': n_init, 'n_jobs': n_jobs})

    cluster_func = globals()[algo]

    cluster_func(**kwargs) # clustering
    if data.shape[0] < 100000 and data.obs[class_label].value_counts().min() == 1:
        new_resol = resolution
        while new_resol > 0.0:
            new_resol -= 0.1
            kwargs['resolution'] = new_resol
            cluster_func(**kwargs)
            if data.obs[class_label].value_counts().min() > 1:
                break
        logger.warning("Reduced resolution from {:.2f} to {:.2f} to avoid clusters of size 1.".format(resolution, new_resol))



import numpy as np
import pandas as pd
import time
from natsort import natsorted

import multiprocessing
from sklearn.cluster import KMeans

from typing import List
from anndata import AnnData

from pegasus.utils import decorators as pg_deco


def estimate_background_probs(adt: AnnData, random_state: int = 0):
    """For cell-hashing data, estimate antibody background probability using EM algorithm.

    Parameters
    ----------
    adt: ``anndata.AnnData``
        Annotated data matrix for antibody.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    Returns
    -------
    ``None``

    Update ``adt.uns``:
        * ``adt.uns["background_probs"]``: estimated antibody background probability.

    Example
    -------
    >>> pg.estimate_background_probs(adt)
    """
    adt.obs["counts"] = adt.X.sum(axis=1).A1 if adt.shape[1] > 1 else adt.X
    counts_log10 = np.log10(adt.obs["counts"].values.reshape(-1, 1))
    kmeans = KMeans(n_clusters=2, random_state=random_state).fit(counts_log10)
    signal = 0 if kmeans.cluster_centers_[0] > kmeans.cluster_centers_[1] else 1
    adt.obs["hto_type"] = "background"
    adt.obs.loc[kmeans.labels_ == signal, "hto_type"] = "signal"

    idx = np.isin(adt.obs["hto_type"], "background")
    pvec = (
        adt.X[idx,].sum(axis=0).A1 if adt.shape[1] > 1 else np.array(adt.X[idx,].sum())
    )
    pvec /= pvec.sum()

    adt.uns["background_probs"] = pvec


def estimate_probs(arr, pvec, alpha, alpha_noise, tol):
    probs = np.zeros(pvec.size + 1)
    old_probs = np.zeros(pvec.size + 1)
    z = np.zeros(pvec.size + 1)
    noise = pvec.size
    # Estimate MLE without Generalized Dirichlet prior
    probs_mle = arr / arr.sum()
    probs[noise] = (probs_mle / pvec).min() + 0.01
    probs[:-1] = np.maximum(probs_mle - probs[noise] * pvec, 0.01)
    probs = probs / probs.sum()

    # EM algorithm
    i = 0
    eps = 1.0
    while eps > tol:
        i += 1
        old_probs[:] = probs[:]
        # E step
        z[:-1] = alpha - 1.0
        z[noise] = alpha_noise - 1.0
        for j in range(pvec.size):
            if arr[j] > 0:
                p = probs[j] / (probs[noise] * pvec[j] + probs[j])
                z[j] += arr[j] * p
                z[noise] += arr[j] * (1.0 - p)
        # M step
        idx = z > 0.0
        probs[idx] = z[idx] / z[idx].sum()
        probs[~idx] = 0.0
        eps = np.linalg.norm(probs - old_probs, ord=1)
        # print ("i = {}, eps = {:.2g}.".format(i, eps))

    return probs


def get_droplet_info(probs, sample_names):
    ids = np.nonzero(probs >= 0.1)[0]
    ids = ids[np.argsort(probs[ids])[::-1]]
    return (
        "singlet" if ids.size == 1 else "doublet",
        ",".join([sample_names[i] for i in ids]),
    )


def calc_demux(data, adt, nsample, min_signal, probs="raw_probs"):
    demux_type = np.full(data.shape[0], "unknown", dtype="object")
    assignments = np.full(data.shape[0], "", dtype="object")

    signals = adt.obs["counts"].reindex(data.obs_names, fill_value=0.0).values * (
        1.0 - data.obsm[probs][:, nsample]
    )
    idx = signals >= min_signal

    tmp = data.obsm[probs][idx,]
    norm_probs = tmp[:, 0:nsample] / (1.0 - tmp[:, nsample])[:, None]

    values1 = []
    values2 = []
    for i in range(norm_probs.shape[0]):
        droplet_type, droplet_id = get_droplet_info(norm_probs[i,], adt.var_names)
        values1.append(droplet_type)
        values2.append(droplet_id)

    demux_type[idx] = values1
    data.obs["demux_type"] = pd.Categorical(
        demux_type, categories=["singlet", "doublet", "unknown"]
    )
    assignments[idx] = values2
    data.obs["assignment"] = pd.Categorical(
        assignments, categories=natsorted(np.unique(assignments))
    )

def has_duplicate_names(names: List[str]) -> bool:
    for name in names:
        if name.find(".#~") >= 0:
            return True
    return False

def remove_suffix(assigns: List[str]) -> List[str]:
    assigns = assigns.astype(str)
    results = []
    for value in assigns:
        fields = value.split(",")
        for i, item in enumerate(fields):
            pos = item.find(".#~")
            if pos >= 0:
                fields[i] = item[:pos]
        results.append(",".join(fields))
    results = np.array(results)
    return pd.Categorical(results, categories=natsorted(np.unique(results)))


@pg_deco.TimeLogger()
def demultiplex(
    data: AnnData,
    adt: AnnData,
    min_signal: float = 10.0,
    alpha: float = 0.0,
    alpha_noise: float = 1.0,
    tol: float = 1e-6,
    n_threads: int = 1,
):
    """Demultiplexing cell-hashing data, using the estimated antibody background probability calculated in ``pg.estimate_background_probs``.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix for gene expression matrix.

    adt: ``anndata.AnnData``
        Annotated data matrix for antibody count matrix.

    min_signal: ``float``, optional, default: ``10.0``
        Any cell/nucleus with less than ``min_signal`` hashtags from the signal will be marked as ``unknown``.

    alpha: ``float``, optional, default: ``0.0``
        The Dirichlet prior concentration parameter (alpha) on samples. An alpha value < 1.0 will make the prior sparse.

    alpha_noise: ``float``, optional, default: ``1.0``

    tol: ``float``, optional, default: ``1e-6``
        Tolerance threshold when judging equivalence of two floating point values.

    n_threads: ``int``, optional, default: ``1``
        Number of threads to use. Must be a positive integer.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs["demux_type"]``: Demultiplexed types of the cells. Either ``singlet``, ``doublet``, or ``unknown``.
        * ``data.obs["assignment"]``: Assigned samples of origin for each cell barcode.
        * ``data.obs["assignment.dedup"]``: Only exist if one sample name can correspond to multiple feature barcodes. In this array, each feature barcode is assigned a unique sample name.

    Examples
    --------
    >>> pg.demultiplex(adata, adt)
    """

    nsample = adt.shape[1]
    data.uns["background_probs"] = adt.uns["background_probs"]

    idx_df = data.obs_names.isin(adt.obs_names)
    adt.obs["rna_type"] = "background"
    adt.obs.loc[data.obs_names[idx_df], "rna_type"] = "signal"

    if nsample == 1:
        print("Warning: detect only one barcode, no need to demultiplex!")
        data.obsm["raw_probs"] = np.zeros((data.shape[0], nsample + 1))
        data.obsm["raw_probs"][:, 0] = 1.0
        data.obsm["raw_probs"][:, 1] = 0.0
        data.obs["demux_type"] = "singlet"
        data.obs["assignment"] = adt.var_names[0]
    else:
        if nsample == 2:
            print(
                "Warning: detect only two barcodes, demultiplexing accuracy might be affected!"
            )

        ncalc = idx_df.sum()
        if ncalc < data.shape[0]:
            nzero = data.shape[0] - ncalc
            print(
                "Warning: {} cells do not have ADTs, percentage = {:.2f}%.".format(
                    nzero, nzero * 100.0 / data.shape[0]
                )
            )
        adt_small = adt[data.obs_names[idx_df],].X.toarray()

        data.obsm["raw_probs"] = np.zeros((data.shape[0], nsample + 1))
        data.obsm["raw_probs"][:, nsample] = 1.0

        iter_array = [
            (adt_small[i,], adt.uns["background_probs"], alpha, alpha_noise, tol)
            for i in range(ncalc)
        ]
        with multiprocessing.Pool(n_threads) as pool:
            data.obsm["raw_probs"][idx_df, :] = pool.starmap(estimate_probs, iter_array)

        calc_demux(data, adt, nsample, min_signal)

        if has_duplicate_names(adt.var_names):
            data.obs["assignment.dedup"] = data.obs["assignment"]
            data.obs["assignment"] = remove_suffix(data.obs["assignment"].values)




#!/usr/bin/env python

import gzip
import logging
import os.path
import re
from typing import List, Tuple, Dict

import anndata
import numpy as np
import pandas as pd
import tables
from scipy.io import mmread
from scipy.sparse import csr_matrix, issparse

from . import Array2D, MemData
from .anndata_legacy import _to_dict_fixed_width_arrays, _parse_whitelist, _update_backed_h5ad

logger = logging.getLogger("pegasus")
from pegasus.utils import decorators as pg_deco


def load_10x_h5_file_v2(h5_in: "tables.File", fn: str, ngene: int = None) -> "MemData":
    """Load 10x v2 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : tables.File
        An instance of tables.File class that is connected to a 10x v2 formatted hdf5 file.
    fn : `str`
        File name, can be used to indicate channel-specific name prefix.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    An MemData object containing genome-Array2D pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v2(h5_in)
    """

    data = MemData()
    for group in h5_in.list_nodes("/", "Group"):
        genome = group._v_name

        M, N = h5_in.get_node("/" + genome + "/shape").read()
        mat = csr_matrix(
            (
                h5_in.get_node("/" + genome + "/data").read(),
                h5_in.get_node("/" + genome + "/indices").read(),
                h5_in.get_node("/" + genome + "/indptr").read(),
            ),
            shape=(N, M),
        )

        barcodes = h5_in.get_node("/" + genome + "/barcodes").read().astype(str)
        ids = h5_in.get_node("/" + genome + "/genes").read().astype(str)
        names = h5_in.get_node("/" + genome + "/gene_names").read().astype(str)

        array2d = Array2D(
            {"barcodekey": barcodes}, {"featurekey": ids, "featurename": names}, mat
        )
        array2d.filter(ngene=ngene)
        array2d.separate_channels(fn)

        data.addData(genome, array2d)

    return data


def load_10x_h5_file_v3(h5_in: "tables.File", fn: str, ngene: int = None) -> "MemData":
    """Load 10x v3 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : tables.File
        An instance of tables.File class that is connected to a 10x v3 formatted hdf5 file.
    fn : `str`
        File name, can be used to indicate channel-specific name prefix.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    An MemData object containing genome-Array2D pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v3(h5_in)
    """

    M, N = h5_in.get_node("/matrix/shape").read()
    bigmat = csr_matrix(
        (
            h5_in.get_node("/matrix/data").read(),
            h5_in.get_node("/matrix/indices").read(),
            h5_in.get_node("/matrix/indptr").read(),
        ),
        shape=(N, M),
    )
    barcodes = h5_in.get_node("/matrix/barcodes").read().astype(str)
    genomes = h5_in.get_node("/matrix/features/genome").read().astype(str)
    ids = h5_in.get_node("/matrix/features/id").read().astype(str)
    names = h5_in.get_node("/matrix/features/name").read().astype(str)

    data = MemData()
    for genome in np.unique(genomes):
        idx = genomes == genome

        barcode_metadata = {"barcodekey": barcodes}
        feature_metadata = {"featurekey": ids[idx], "featurename": names[idx]}
        mat = bigmat[:, idx].copy()
        array2d = Array2D(barcode_metadata, feature_metadata, mat)
        array2d.filter(ngene)
        array2d.separate_channels(fn)

        data.addData(genome, array2d)

    return data


def load_10x_h5_file(input_h5: str, ngene: int = None) -> "MemData":
    """Load 10x format matrix (either v2 or v3) from hdf5 file

    Parameters
    ----------

    input_h5 : `str`
        The matrix in 10x v2 or v3 hdf5 format.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    An MemData object containing genome-Array2D pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file('example_10x.h5')
    """

    fn = os.path.basename(input_h5)[:-3]

    data = None
    with tables.open_file(input_h5) as h5_in:
        try:
            node = h5_in.get_node("/matrix")
            data = load_10x_h5_file_v3(h5_in, fn, ngene)
        except tables.exceptions.NoSuchNodeError:
            data = load_10x_h5_file_v2(h5_in, fn, ngene)

    return data


def _enumerate_files(path: str, parts: List[str], repl_list1: List[str], repl_list2: List[str] = None) -> str:
    """ Enumerate all possible file names """
    if len(parts) <= 2:
        for token in repl_list1:
            parts[-1] = token
            candidate = os.path.join(path, ''.join(parts))
            if os.path.isfile(candidate):
                return candidate
    else:
        assert len(parts) == 4
        for p2 in repl_list1:
            parts[1] = p2
            for p4 in repl_list2:
                parts[3] = p4
                candidate = os.path.join(path, ''.join(parts))
                if os.path.isfile(candidate):
                    return candidate
    return None


def _locate_barcode_and_feature_files(path: str, fname: str) -> Tuple[str, str]:
    """ Locate barcode and feature files (with path) based on mtx file name (no suffix)
    """
    barcode_file = feature_file = None
    if fname == "matrix":
        barcode_file = _enumerate_files(path, [''], ["cells.tsv.gz", "cells.tsv", "barcodes.tsv.gz", "barcodes.tsv"])
        feature_file = _enumerate_files(path, [''], ["genes.tsv.gz", "genes.tsv", "features.tsv.gz", "features.tsv"])
    else:
        p1, p2, p3 = fname.partition("matrix")
        if p2 == '' and p3 == '':
            barcode_file = _enumerate_files(path, [p1, ''], [".barcodes.tsv.gz", ".barcodes.tsv", ".cells.tsv.gz", ".cells.tsv", "_barcode.tsv", ".barcodes.txt"])
            feature_file = _enumerate_files(path, [p1, ''], [".genes.tsv.gz", ".genes.tsv", ".features.tsv.gz", ".features.tsv", "_gene.tsv", ".genes.txt"])
        else:
            barcode_file = _enumerate_files(path, [p1, '', p3, ''], ["barcodes", "cells"], [".tsv.gz", ".tsv"])
            feature_file = _enumerate_files(path, [p1, '', p3, ''], ["genes", "features"], [".tsv.gz", ".tsv"])

    if barcode_file is None:
        raise ValueError("Cannot find barcode file!")
    if feature_file is None:
        raise ValueError("Cannot find feature file!")

    return barcode_file, feature_file


def _load_barcode_metadata(barcode_file: str) -> "pd.DataFrame":
    """ Load cell barcode information """
    barcode_metadata = pd.read_csv(barcode_file, sep="\t", header=None)
    if "cellkey" in barcode_metadata.iloc[0].values:
        # HCA DCP format
        barcode_metadata = pd.DataFrame(data = barcode_metadata.iloc[1:].values, columns = barcode_metadata.iloc[0])
        barcode_metadata.rename(columns={"cellkey": "barcodekey"}, inplace=True)
    else:
        # Other format, only one column containing the barcode is expected
        barcode_metadata = pd.read_csv(barcode_file, sep="\t", header=None, names=["barcodekey"])

    return barcode_metadata


def _load_feature_metadata(feature_file: str) -> Tuple["pd.DataFrame", str]:
    """ Load feature information """
    feature_metadata = pd.read_csv(feature_file, sep="\t", header=None)
    format_type = ""
    if "featurekey" in feature_metadata.iloc[0].values:
        # HCA DCP format genes.tsv.gz
        feature_metadata = pd.DataFrame(data = feature_metadata.iloc[1:].values, columns = feature_metadata.iloc[0])
        format_type = "HCA DCP"
    else:
        assert feature_metadata.shape[1] <= 3
        col_names = ["featurekey", "featurename", "featuretype"]
        # shape[1] = 3 -> 10x v3, 2 -> 10x v2, 1 -> scumi, dropEst, BUStools
        feature_metadata.columns = col_names[:feature_metadata.shape[1]]
        if feature_metadata.shape[1] == 3:
            format_type = "10x v3"
        elif feature_metadata.shape[1] == 2:
            format_type = "10x v2"
        else:
            assert feature_metadata.shape[1] == 1
            # Test if in scumi format
            if feature_metadata.iloc[0, 0].find('_') >= 0:
                format_type = "scumi"
                values = feature_metadata.values[:, 0].astype(str)
                arr = np.array(np.char.split(values, sep="_", maxsplit=1).tolist())
                feature_metadata = pd.DataFrame(data={"featurekey": arr[:, 0], "featurename": arr[:, 1]})
            else:
                format_type = "dropEst or BUStools"
                feature_metadata["featurename"] = feature_metadata["featurekey"]            

    return feature_metadata, format_type


def load_one_mtx_file(path: str, file_name: str, ngene: int = None) -> "Array2D":
    """Load one gene-count matrix in mtx format into an Array2D object
    """
    mtx_file = os.path.join(path, file_name)
    mat = mmread(mtx_file)

    fname = re.sub('(.mtx|.mtx.gz)$', '', file_name)
    barcode_file, feature_file = _locate_barcode_and_feature_files(path, fname)

    barcode_metadata = _load_barcode_metadata(barcode_file)
    feature_metadata, format_type = _load_feature_metadata(feature_file)
    logger.info("Detected mtx file in {} format.".format(format_type))
    
    if mat.shape[1] == barcode_metadata.shape[0]: # Column is barcode, transpose the matrix
        mat = mat.T

    array2d = Array2D(barcode_metadata, feature_metadata, csr_matrix(mat))
    array2d.filter(ngene=ngene)
    if format_type == "10x v3" or format_type == "10x v2":
        array2d.separate_channels("")  # fn == '' refers to 10x mtx format

    return array2d


def _locate_mtx_file(path: str) -> str:
    """ Locate one mtx file in the path directory """
    with os.scandir(path) as dir_iter:
        for dir_entry in dir_iter:
            file_name = dir_entry.name
            if file_name.endswith('.mtx.gz') or file_name.endswith('.mtx'):
                return file_name
    return None


def load_mtx_file(path: str, genome: str = None, ngene: int = None) -> "MemData":
    """Load gene-count matrix from Market Matrix files (10x v2, v3 and HCA DCP formats)

    Parameters
    ----------

    path : `str`
        Path to mtx files. The directory implied by path should either contain matrix, feature and barcode information, or folders containing these information.
    genome : `str`, optional (default: None)
        Genome name of the matrix. If None, genome will be inferred from path.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    An MemData object containing a genome-Array2D pair.

    Examples
    --------
    >>> io.load_mtx_file('example.mtx.gz', genome = 'mm10')
    """

    orig_file = path
    if os.path.isdir(orig_file):
        path = orig_file.rstrip('/')
        file_name = _locate_mtx_file(path)
    else:
        if (not orig_file.endswith(".mtx")) and (not orig_file.endswith(".mtx.gz")):
            raise ValueError("File {} does not end with suffix .mtx or .mtx.gz!".format(orig_file))
        path, file_name = os.path.split(orig_file)

    data = MemData()

    if file_name is not None:
        if genome is None:
            genome = os.path.basename(path)
        data.addData(
            genome,
            load_one_mtx_file(
                path,
                file_name,
                ngene=ngene,
            ),
        )
    else:
        for dir_entry in os.scandir(path):
            if dir_entry.is_dir():
                file_name = _locate_mtx_file(dir_entry.path)
                if file_name is None:
                    raise ValueError("Folder {} does not contain a mtx file!".format(dir_entry.path))
                data.addData(dir_entry.name, load_one_mtx_file(dir_entry.path, file_name, ngene=ngene))

    return data


def _load_csv_file_sparse(input_csv, genome, sep, dtype, ngene=None, chunk_size=1000):
    """
     Read a csv file in chunks
    """

    import scipy.sparse
    features = []
    dense_arrays = []
    sparse_arrays = []

    if chunk_size <= 0:
        raise ValueError("Chunk size must be greater than zero")

    with (
        gzip.open(input_csv, mode="rt")
        if input_csv.endswith(".gz")
        else open(input_csv)
    ) as fin:
        barcodes = next(fin).strip().split(sep)[1:]
        for line in fin:
            fields = line.strip().split(sep)
            features.append(fields[0])
            dense_arrays.append(np.array(fields[1:], dtype=dtype))
            if len(dense_arrays) == chunk_size:
                sparse_arrays.append(
                    scipy.sparse.csr_matrix(np.stack(dense_arrays, axis=0))
                )
                dense_arrays = []

    if len(dense_arrays) > 0:
        sparse_arrays.append(scipy.sparse.csr_matrix(np.stack(dense_arrays, axis=0)))
        dense_arrays = None
    mat = scipy.sparse.vstack(sparse_arrays)
    barcode_metadata = {"barcodekey": barcodes}
    feature_metadata = {"featurekey": features, "featurename": features}
    data = MemData()
    array2d = Array2D(barcode_metadata, feature_metadata, mat.T)
    array2d.filter(ngene=ngene)
    data.addData(genome, array2d)
    return data


def load_csv_file(
    input_csv: str,
    genome: str,
    sep: str = ",",
    ngene: int = None,
    chunk_size: int = None,
) -> "MemData":
    """Load count matrix from a CSV-style file, such as CSV file or DGE style tsv file.

     Parameters
     ----------

     input_csv : `str`
         The CSV file, gzipped or not, containing the count matrix.
     genome : `str`
         The genome reference.
     sep: `str`, optional (default: ',')
         Separator between fields, either ',' or '\t'.
     ngene : `int`, optional (default: None)
         Minimum number of genes to keep a barcode. Default is to keep all barcodes.
     chunk_size: `int`, optional (default: None)
        Chunk size for reading dense matrices as sparse
     Returns
     -------

     An MemData object containing a genome-Array2D pair.

     Examples
     --------
     >>> io.load_csv_file('example_ADT.csv', genome = 'GRCh38')
     >>> io.load_csv_file('example.umi.dge.txt.gz', genome = 'GRCh38', sep = '\t')
     """

    path = os.path.dirname(input_csv)
    base = os.path.basename(input_csv)
    is_hca_csv = base == "expression.csv"

    if sep == "\t":
        # DGE, columns are cells, which is around thousands and we can use pandas.read_csv
        if chunk_size is not None:
            return _load_csv_file_sparse(
                input_csv,
                genome,
                sep,
                "float32" if base.startswith("expression") else "int",
                ngene=ngene,
                chunk_size=chunk_size)
        df = pd.read_csv(input_csv, header=0, index_col=0, sep=sep)
        mat = csr_matrix(df.values.T)
        barcode_metadata = {"barcodekey": df.columns.values}
        feature_metadata = {
            "featurekey": df.index.values,
            "featurename": df.index.values,
        }
    else:
        if chunk_size is not None and not is_hca_csv:
            return _load_csv_file_sparse(
                input_csv,
                genome,
                sep,
                "float32" if base.startswith("expression") else "int",
                ngene=ngene,
                chunk_size=chunk_size,
            )
        # For CSV files, wide columns prevent fast pd.read_csv loading
        converter = (
            float if base.startswith("expression") else int
        )  # If expression -> float otherwise int

        barcodes = []
        names = []
        stacks = []
        with (
            gzip.open(input_csv, mode="rt")
            if input_csv.endswith(".gz")
            else open(input_csv)
        ) as fin:
            barcodes = next(fin).strip().split(sep)[1:]
            for line in fin:
                fields = line.strip().split(sep)
                names.append(fields[0])
                stacks.append([converter(x) for x in fields[1:]])

        mat = csr_matrix(np.stack(stacks, axis=1 if not is_hca_csv else 0))
        barcode_metadata = {"barcodekey": barcodes}
        feature_metadata = {"featurekey": names, "featurename": names}

        if is_hca_csv:
            barcode_file = os.path.join(path, "cells.csv")
            if os.path.exists(barcode_file):
                barcode_metadata = pd.read_csv(barcode_file, sep=",", header=0)
                assert "cellkey" in barcode_metadata
                barcode_metadata.rename(columns={"cellkey": "barcodekey"}, inplace=True)

            feature_file = os.path.join(path, "genes.csv")
            if os.path.exists(feature_file):
                feature_metadata = pd.read_csv(feature_file, sep=",", header=0)

    data = MemData()
    array2d = Array2D(barcode_metadata, feature_metadata, mat)
    array2d.filter(ngene=ngene)
    data.addData(genome, array2d)

    return data


def load_loom_file(input_loom: str, genome: str, ngene: int = None) -> "MemData":
    """Load count matrix from a LOOM file. Currently only support HCA DCP Loom spec.

    Parameters
    ----------

    input_loom : `str`
        The LOOM file, containing the count matrix.
    genome : `str`
        The genome reference.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    An MemData object containing a genome-Array2D pair.

    Examples
    --------
    >>> io.load_loom_file('example.loom', genome = 'GRCh38', ngene = 200)
    """
    import loompy

    ds = loompy.connect(input_loom)
    
    if "obs_names" in ds.col_attrs.keys():
        # If loom is generated by anndata
        ds.close()
        data = anndata.read_loom(input_loom)
        return data
    
    col_trans = {"CellID": "barcodekey"}
    row_trans = {"Accession": "featurekey", "Gene": "featurename"}

    data = MemData()
    mat = csr_matrix(ds.sparse().T)
    barcode_metadata = {}
    for keyword, values in ds.col_attrs.items():
        keyword = col_trans.get(keyword, keyword)
        barcode_metadata[keyword] = values
    feature_metadata = {}
    for keyword, values in ds.row_attrs.items():
        keyword = row_trans.get(keyword, keyword)
        feature_metadata[keyword] = values
    ds.close()

    array2d = Array2D(barcode_metadata, feature_metadata, mat)
    array2d.filter(ngene=ngene)
    data.addData(genome, array2d)

    return data


def load_pegasus_h5_file(
    input_h5: str, ngene: int = None, select_singlets: bool = False
) -> "MemData":
    """Load matrices from pegasus-format hdf5 file

    Parameters
    ----------

    input_h5 : `str`
        pegasus-format hdf5 file.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.
    select_singlets: `bool`, optional (default: False)
        If only load singlets.

    Returns
    -------

    An MemData object containing genome-Array2D pair per genome.

    Examples
    --------
    >>> io.load_pegasus_h5_file('example.h5sc')
    """

    cite_seq_name = None
    selected_barcodes = None

    data = MemData()
    with tables.open_file(input_h5) as h5_in:
        for group in h5_in.list_nodes("/", "Group"):
            genome = group._v_name

            M, N = h5_in.get_node("/" + genome + "/shape").read()
            mat = csr_matrix(
                (
                    h5_in.get_node("/" + genome + "/data").read(),
                    h5_in.get_node("/" + genome + "/indices").read(),
                    h5_in.get_node("/" + genome + "/indptr").read(),
                ),
                shape=(N, M),
            )

            barcode_metadata = {}
            for node in h5_in.walk_nodes("/" + genome + "/_barcodes", "Array"):
                values = node.read()
                if values.dtype.kind == "S":
                    values = values.astype(str)
                barcode_metadata[node.name] = values

            feature_metadata = {}
            for node in h5_in.walk_nodes("/" + genome + "/_features", "Array"):
                values = node.read()
                if values.dtype.kind == "S":
                    values = values.astype(str)
                feature_metadata[node.name] = values

            array2d = Array2D(barcode_metadata, feature_metadata, mat)
            if genome.startswith("CITE_Seq"):
                cite_seq_name = genome
            else:
                array2d.filter(ngene, select_singlets)
                selected_barcodes = array2d.get_metadata("barcodekey")

            data.addData(genome, array2d)

    if (cite_seq_name is not None) and (selected_barcodes is not None):
        array2d = data.getData(cite_seq_name)
        selected = array2d.get_metadata("barcodekey").isin(selected_barcodes)
        array2d.trim(selected)

    return data


def infer_file_format(input_file: str) -> Tuple[str, str, str]:
    """ Infer file format from input_file name

    This function infer file format by inspecting the file name.

    Parameters
    ----------

    input_file : `str`
        Input file name.

    Returns
    -------
    `str`
        File format, choosing from '10x', 'pegasus', 'h5ad', 'loom', 'mtx', 'dge', 'csv' and 'tsv'.
    `str`
        The path covering all input files. Most time this is the same as input_file. But for HCA mtx and csv, this should be parent directory.
    `str`
        Type of the path, either 'file' or 'directory'.
    """

    file_format = None
    copy_path = input_file
    copy_type = "file"

    if input_file.endswith(".h5"):
        file_format = "10x"
    elif input_file.endswith(".h5sc"):
        file_format = "pegasus"
    elif input_file.endswith(".h5ad"):
        file_format = "h5ad"
    elif input_file.endswith(".loom"):
        file_format = "loom"
    elif (
        input_file.endswith(".mtx")
        or input_file.endswith(".mtx.gz")
        or os.path.splitext(input_file)[1] == ""
    ):
        file_format = "mtx"
        if os.path.splitext(input_file)[1] != "":
            copy_path = os.path.dirname(input_file)
        copy_type = "directory"
    elif input_file.endswith("dge.txt.gz"):
        file_format = "dge"
    elif input_file.endswith(".csv") or input_file.endswith(".csv.gz"):
        file_format = "csv"
        if os.path.basename(input_file) == "expression.csv":
            copy_path = os.path.dirname(input_file)
            copy_type = "directory"
    elif input_file.endswith(".txt") or input_file.endswith(".tsv") or input_file.endswith(
        ".txt.gz") or input_file.endswith(".tsv.gz"):
        file_format = "tsv"
    else:
        raise ValueError("Unrecognized file type for file {}!".format(input_file))

    return file_format, copy_path, copy_type


@pg_deco.TimeLogger()
def read_input(
    input_file: str,
    genome: str = None,
    return_type: str = "AnnData",
    concat_matrices: bool = False,
    h5ad_mode: str = "a",
    ngene: int = None,
    select_singlets: bool = False,
    channel_attr: str = None,
    chunk_size: int = None,
    black_list: List[str] = [],
) -> "MemData or AnnData or List[AnnData]":
    """Load data into memory.

    This function is used to load input data into memory. Inputs can be in 10x genomics v2 & v3 formats (hdf5 or mtx), HCA DCP mtx and csv formats, Drop-seq dge format, and CSV format.

    Parameters
    ----------

    input_file : `str`
        Input file name.
    genome : `str`, optional (default: None)
        A string contains comma-separated genome names. pegasus will read all matrices matching the genome names. If genome is None, all matrices will be considered. For formats like loom, mtx, dge, csv and tsv, genome is used to provide genome name. In this case if genome is None, except mtx format, '' is used as the genome name instead.
    return_type : `str`
        Return object type, can be either 'MemData' or 'AnnData'.
    concat_matrices : `boolean`, optional (default: False)
        If input file contains multiple matrices, turning this option on will concatenate them into one AnnData object. Otherwise return a list of AnnData objects.
    h5ad_mode : `str`, optional (default: 'a')
        If input is in h5ad format, the backed mode for loading the data. Mode could be 'a', 'r', 'r+', where 'a' refers to load the whole matrix into memory.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.
    select_singlets : `bool`, optional (default: False)
        If this option is on, only keep DemuxEM-predicted singlets when loading data.
    channel_attr : `str`, optional (default: None)
        Use channel_attr to represent different samples. This will set a 'Channel' column field with channel_attr.
    chunk_size: `int`, optional (default: None)
        Chunk size for reading dense matrices as sparse
    black_list : `List[str]`, optional (default: [])
        Attributes in black list will be poped out.

    Returns
    -------
    `MemData` object or `anndata` object or a list of `anndata` objects
        An `MemData` object or `anndata` object or a list of `anndata` objects containing the count matrices.

    Examples
    --------
    >>> adata = pg.read_input('example_10x.h5', genome = 'mm10')
    >>> adata = pg.read_input('example.h5ad', h5ad_mode = 'r+')
    >>> adata = pg.read_input('example_ADT.csv')
    """

    input_file = os.path.expanduser(os.path.expandvars(input_file))
    file_format, _, _ = infer_file_format(input_file)

    if file_format == "pegasus":
        data = load_pegasus_h5_file(
            input_file, ngene=ngene, select_singlets=select_singlets
        )
    elif file_format == "10x":
        data = load_10x_h5_file(input_file, ngene=ngene)
    elif file_format == "h5ad":
        data = anndata.read_h5ad(
            input_file,
            chunk_size=chunk_size,
            backed=(None if h5ad_mode == "a" else h5ad_mode),
        )
        if channel_attr is not None:
            data.obs["Channel"] = data.obs[channel_attr]
        for attr in black_list:
            if attr in data.obs:
                data.obs.drop(columns = attr, inplace = True)
    elif file_format == "mtx":
        data = load_mtx_file(input_file, genome, ngene=ngene)
    elif file_format == "loom":
        if genome is None:
            genome = ''
        data = load_loom_file(input_file, genome, ngene=ngene)
        if isinstance(data, anndata.AnnData):
            file_format = "h5ad"
            if channel_attr is not None:
                data.obs["Channel"] = data.obs[channel_attr]
            for attr in black_list:
                if attr in data.obs:
                    data.obs.drop(columns = attr, inplace = True)
    else:
        assert file_format == "dge" or file_format == "csv" or file_format == "tsv"
        if genome is None:
            genome = ''
        data = load_csv_file(
            input_file,
            genome,
            sep=("\t" if file_format == "dge" or file_format == "tsv" else ","),
            ngene=ngene,
            chunk_size=chunk_size,
        )

    if file_format != "h5ad":
        data.restrain_keywords(genome)
        if return_type == "AnnData":
            data = data.convert_to_anndata(
                concat_matrices=concat_matrices,
                channel_attr=channel_attr,
                black_list=black_list,
            )
    else:
        assert return_type == "AnnData"
        if select_singlets:
            assert 'demux_type' in data.obs
            data._inplace_subset_obs((data.obs['demux_type'] == 'singlet').values)

    return data


def _write_mtx(data: "AnnData", output_file: str):
    import scipy.io
    import gzip
    import shutil
    output_dir = os.path.dirname(os.path.abspath(output_file))
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    mtx_file = os.path.join(output_dir, 'matrix.mtx')
    scipy.io.mmwrite(mtx_file, data.X.T)
    mtx_file_gz = "%s.gz" % mtx_file

    with open(mtx_file, 'rb') as f_in:
        with gzip.open(mtx_file_gz, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(mtx_file)

    data.obs.to_csv(os.path.join(output_dir, 'barcodes.tsv.gz'), header=None, columns=[])
    features_df = pd.DataFrame()
    if 'gene_ids' in data.var:
        features_df['gene_ids'] = data.var['gene_ids']
        features_df['gene_names'] = data.var.index
    else:
        features_df['gene_ids'] = data.var.index
        features_df['gene_names'] = data.var.index
    features_df['type'] = 'Gene Expression'
    features_df.to_csv(os.path.join(output_dir, 'features.tsv.gz'), header=None, sep='\t', index=False)
    data.obs.to_csv(os.path.join(output_dir, 'obs.csv.gz'), index_label='id')
    var_columns = list(data.var.columns)
    if 'gene_ids' in data.var:
        var_columns.remove('gene_ids')
    if len(var_columns) > 0:
        data.var.to_csv(os.path.join(output_dir, 'var.csv.gz'), index_label='id', columns=var_columns)


def _write_loom(data: "AnnData", output_file: str):
    
    def _process_attrs(key_name: str, attrs: pd.DataFrame, attrs_multi: 'anndata._core.aligned_mapping.AxisArrays') -> Dict[str, object]:
        res_dict = {key_name: attrs.index.values}
        for key in attrs.columns:
            res_dict[key] = np.array(attrs[key].values)
        for key in attrs_multi.keys():
            value = attrs_multi[key]
            if value.ndim > 1:
                res_dict[key] = value if value.shape[1] > 1 else value[:, 0]
        return res_dict

    row_attrs = _process_attrs('var_names', data.var, data.varm)
    col_attrs = _process_attrs('obs_names', data.obs, data.obsm)

    import loompy
    loompy.create(output_file, data.X.T, row_attrs = row_attrs, col_attrs = col_attrs)
    logger.info('{0} is written.'.format(output_file))


@pg_deco.TimeLogger()
def write_output(
    data: "MemData or AnnData",
    output_file: str,
    whitelist: List = ["obs", "obsm", "uns", "var", "varm"],
) -> None:
    """ Write data back to disk.

    This function is used to write data back to disk.

    Parameters
    ----------

    data : `MemData` or `AnnData`
        data to write back, can be either an MemData or AnnData object.
    output_file : `str`
        output file name. If data is MemData, output_file should ends with suffix '.h5sc'. Otherwise, output_file can end with either '.h5ad', '.loom', or '.mtx.gz'. If output_file ends with '.loom', a LOOM file will be generated. If no suffix is detected, an appropriate one will be appended.
    whitelist : `list`, optional, default = ["obs", "obsm", "uns", "var", "varm"]
        List that indicates changed fields when writing h5ad file in backed mode. For example, ['uns/Groups', 'obsm/PCA'] will only write Groups in uns, and PCA in obsm; the rest of the fields will be unchanged.

    Returns
    -------
    `None`

    Examples
    --------
    >>> pg.write_output(adata, 'test.h5ad')
    """

    if (not isinstance(data, MemData)) and (not isinstance(data, anndata.AnnData)):
        raise ValueError("data is neither an MemData nor AnnData object!")

    # Identify and correct file suffix
    file_name, _, suffix = output_file.rpartition(".")
    if suffix == 'gz' and file_name.endswith('.mtx'):
        suffix = 'mtx.gz'
    if file_name == "":
        file_name = output_file
        suffix = "h5sc" if isinstance(data, MemData) else "h5ad"
    if isinstance(data, MemData) and suffix != "h5sc" and suffix != "h5":
        logging.warning(
            "Detected file suffix for this MemData object is neither .h5sc nor .h5. We will assume output_file is a file name and append .h5sc suffix."
        )
        file_name = output_file
        suffix = "h5sc"
    if isinstance(data, anndata.AnnData) and (suffix not in ["h5ad", "loom", "mtx.gz"]):
        logging.warning(
            "Detected file suffix for this AnnData object is neither .h5ad or .loom. We will assume output_file is a file name and append .h5ad suffix."
        )
        file_name = output_file
        suffix = "h5ad"
    output_file = file_name + "." + suffix

    # Eliminate objects starting with fmat_ from uns
    if isinstance(data, anndata.AnnData):
        keys = list(data.uns)
        for keyword in keys:
            if keyword.startswith("fmat_"):
                data.uns.pop(keyword)

    # Write outputs
    if suffix == "mtx.gz":
        _write_mtx(data, output_file)
    elif suffix == "h5sc" or suffix == "h5":
        data.write_h5_file(output_file)
    elif suffix == "loom":
        _write_loom(data, output_file)
    elif (
        not data.isbacked
        or (data.isbacked and data.file._file.mode != "r+")
    ):
        data.write(output_file, compression="gzip")
    else:
        assert data.file._file.mode == "r+"
        import h5py

        h5_file = data.file._file
        # Fix old h5ad files in which obsm/varm were stored as compound datasets
        for key in ["obsm", "varm"]:
            if key in h5_file.keys() and isinstance(h5_file[key], h5py.Dataset):
                del h5_file[key]
                whitelist.append(key)

        _update_backed_h5ad(
            h5_file, _to_dict_fixed_width_arrays(data), _parse_whitelist(whitelist)
        )
        h5_file.close()



import numpy as np
import pandas as pd
import json
import logging

from sys import stdout
from natsort import natsorted
from typing import List, Dict, Union
from anndata import AnnData

logger = logging.getLogger("pegasus")
from pegasus.utils import decorators as pg_deco


class CellType:
    def __init__(self, name: str, ignore_nonde: bool = False):
        self.name = name
        self.score = self.avgp = 0.0
        self.weak_support = []
        self.strong_support = []
        self.subtypes = None
        self.ignore_nonde = ignore_nonde

    def evaluate(
        self,
        obj: "json object",
        de_up: pd.DataFrame,
        de_down: pd.DataFrame,
        thre: float,
    ):
        """ Calculate score for matching a cluster with a putative cell type.
        """
        self.score = self.avgp = 0.0
        self.weak_support = []
        self.strong_support = []

        nump = 0
        for marker_set in obj["markers"]:
            numer = 0.0
            denom = len(marker_set["genes"]) * 2.0
            if denom == 0.0:
                continue

            for marker in marker_set["genes"]:
                sign = marker[-1]
                gsym = marker[:-1]

                if sign == "+":
                    if gsym in de_up.index:
                        fc = de_up.at[gsym, "fc"]
                        percent = de_up.at[gsym, "percent"]
                        self.avgp += percent
                        nump += 1

                        if fc >= thre:
                            numer += 2.0
                            self.strong_support.append(
                                (marker, "{0:.2f}%".format(percent))
                            )
                        else:
                            numer += 1.0 + (fc - 1.0) / (thre - 1.0)
                            self.weak_support.append(
                                (marker, "{0:.2f}%".format(percent))
                            )
                else:
                    assert sign == "-"
                    if gsym not in de_up.index:
                        if gsym in de_down.index:
                            fc = (
                                (1.0 / de_down.at[gsym, "fc"])
                                if de_down.at[gsym, "fc"] > 0.0
                                else np.inf
                            )
                            percent = de_down.at[gsym, "percent"]
                            if fc >= thre:
                                numer += 2.0
                                self.strong_support.append(
                                    (marker, "{0:.2f}%".format(percent))
                                )
                            else:
                                numer += 1.0 + (fc - 1.0) / (thre - 1.0)
                                self.weak_support.append(
                                    (marker, "{0:.2f}%".format(percent))
                                )
                        elif not self.ignore_nonde:
                            numer += 1.0
                            self.weak_support.append((marker, "N/A"))

            self.score += numer / denom * marker_set["weight"]

        self.score = (
            self.score / obj["denominator"] if obj["denominator"] > 0.0 else 0.0
        )
        if nump > 0:
            self.avgp /= nump

    def __repr__(self):
        res = "name: {0}; score: {1:.2f}; average marker percentage: {2:.2f}%".format(
            self.name, self.score, self.avgp
        )
        if len(self.strong_support) > 0:
            res += "; strong support: {0}".format(
                ",".join(["({0},{1})".format(x[0], x[1]) for x in self.strong_support])
            )
        if len(self.weak_support) > 0:
            res += "; weak support: {0}".format(
                ",".join(["({0},{1})".format(x[0], x[1]) for x in self.weak_support])
            )

        return res


class Annotator:
    def __init__(self, marker_file: Union[str, Dict], genes: List[str]) -> None:
        if type(marker_file) != dict:
            with open(marker_file) as fin:
                self.object = json.load(fin)
        else:
            self.object = marker_file
        self.recalibrate(self.object, genes)

    def recalibrate(self, obj: "json object", genes: List[str]) -> None:
        """ Remove markers that are not expressed (not in genes) and calculate partial weights for existing genes.
        """
        for celltype in obj["cell_types"]:
            denom = 0.0
            for marker_set in celltype["markers"]:
                markers = marker_set["genes"]
                s = len(markers)
                marker_set["genes"] = [x for x in markers if x[:-1] in genes]
                new_s = len(marker_set["genes"])
                marker_set["weight"] = marker_set["weight"] / s * new_s
                denom += marker_set["weight"]
            celltype["denominator"] = denom
            sub_obj = celltype.get("subtypes", None)
            if sub_obj is not None:
                self.recalibrate(sub_obj, genes)

    def evaluate(
        self,
        de_up: pd.DataFrame,
        de_down: pd.DataFrame,
        thre: float = 1.5,
        ignore_nonde: bool = False,
        obj: "json object" = None,
    ):
        """ Evaluate a cluster to determine its putative cell type.
        """
        if obj is None:
            obj = self.object

        results = []
        for celltype in obj["cell_types"]:
            ct = CellType(celltype["name"], ignore_nonde=ignore_nonde)
            ct.evaluate(celltype, de_up, de_down, thre)
            if ct.score >= 0.5:
                sub_obj = celltype.get("subtypes", None)
                if sub_obj is not None:
                    ct.subtypes = self.evaluate(
                        de_up,
                        de_down,
                        thre=thre,
                        ignore_nonde=ignore_nonde,
                        obj=sub_obj,
                    )
            results.append(ct)

        results.sort(key=lambda x: x.score, reverse=True)

        return results

    def report(
        self,
        fout: "output stream",
        ct_list: List["CellType"],
        thre: float,
        space: int = 4,
    ) -> None:
        """ Write putative cell type reports to fout.
        """
        for ct in ct_list:
            if ct.score >= thre:
                fout.write(" " * space + str(ct) + "\n")
                if ct.subtypes is not None:
                    self.report(fout, ct.subtypes, 0.5, space + 4)


def infer_cluster_names(
    cell_type_dict: Dict[str, List["CellType"]], threshold: float = 0.5
) -> List[str]:
    """Decide cluster names based on cell types automatically.

    Parameters
    ----------
    cell_type_dict: ``Dict[str, List["CellType"]]``
        Python dictionary of cell type list for each cluster. This is the output of ``pg.infer_cell_types``.

    threshold: ``float``, optional, default: ``0.5``
        A threshold for cell type result reported. It should be a real number between ``0.0`` and ``1.0``.

    Returns
    -------
    ``List[str]``
        A list of cluster names decided by their corresponding cell types. The order is consistent with clusters.

    Examples
    --------
    >>> cell_type_dict = pg.infer_cell_types(adata, markers = 'human_immune', de_test = 't')
    >>> cluster_names = pg.infer_cluster_names(cell_type_dict)
    """
    cluster_ids = natsorted(cell_type_dict.keys())
    names = []
    name_dict = dict()
    for cluster_id in cluster_ids:
        ct_list = cell_type_dict[cluster_id]

        if len(ct_list) == 0 or ct_list[0].score < threshold:
            cell_name = cluster_id
        else:
            ct = ct_list[0]
            while ct.subtypes is not None and ct.subtypes[0].score >= threshold:
                ct = ct.subtypes[0]
            cell_name = ct.name

            if cell_name in name_dict:
                name_dict[cell_name] += 1
                cell_name = cell_name + "-" + str(name_dict[cell_name])
            else:
                name_dict[cell_name] = 1

        names.append(cell_name)

    return names


def infer_cell_types(
    data: AnnData,
    markers: Union[str, Dict],
    de_test: str,
    de_alpha: float = 0.05,
    de_key: str = "de_res",
    threshold: float = 0.5,
    ignore_nonde: bool = False,
    output_file: str = None,
) -> Dict[str, List["CellType"]]:
    """Infer putative cell types for each cluster using legacy markers.

    Parameters
    ----------

    data : ``anndata.AnnData``
        AnnData object of count matrix and DE analysis results.

    markers : ``str`` or ``Dict``
        * If ``str``, it 
            * either refers to a JSON file containing legacy markers, or
            * ``'human_immune'`` for predefined pegasus markers on human immune cells;
            * ``'mouse_immune'`` for mouse immune cells;
            * ``'human_brain'`` for human brain cells;
            * ``'mouse_brain'`` for mouse brain cells.
        * If ``Dict``, it refers to a Python dictionary describing the markers.

    de_test: ``str``
        pegasus determines cell types using DE test results. This argument indicates which DE test result to use, can be either ``'t'``, ``'fisher'`` or ``'mwu'``.

    de_alpha: ``float``, optional, default: ``0.05``
        False discovery rate for controling family-wide error.

    de_key : ``str``, optional, default: ``"de_res"``
        The keyword in ``data.varm`` that stores DE analysis results.

    threshold : ``float``, optional, defaut: ``0.5``
        Only report putative cell types with a score larger than or equal to ``threshold``.

    ignore_nonde: ``bool``, optional, default: ``False``
        Do not consider non DE genes as weak negative markers.

    output_file: ``str``, optional, default: ``None``
        File name of output cluster annotation. If ``None``, do not write to any file.

    Returns
    -------
    ``Dict[str, List["CellType"]]``
        Python dictionary with cluster ID's being keys, and their corresponding cell type lists sortec by scores being values.

    Examples
    --------
    >>> cell_type_dict = pg.infer_cell_types(adata, markers = 'human_immune', de_test = 't')
    """

    if output_file is not None:
        fout = open(output_file, "w")

    predefined_markers = dict(
        human_immune="human_immune_cell_markers.json",
        mouse_immune="mouse_immune_cell_markers.json",
        mouse_brain="mouse_brain_cell_markers.json",
        human_brain="human_brain_cell_markers.json",
    )

    if markers in predefined_markers:
        import pkg_resources

        markers = pkg_resources.resource_filename(
            "pegasus.annotate_cluster", predefined_markers[markers]
        )

    anno = Annotator(markers, data.var_names)

    clusts = natsorted(
        [
            x.rpartition(":")[2]
            for x in data.varm[de_key].dtype.names
            if x.startswith("WAD_score:")
        ]
    )
    cell_type_results = {}
    for clust_id in clusts:
        idx = data.varm[de_key]["{0}_qval:{1}".format(de_test, clust_id)] <= de_alpha

        idx_up = idx & (data.varm[de_key]["log_fold_change:{0}".format(clust_id)] > 0.0)
        idx_down = idx & (
            data.varm[de_key]["log_fold_change:{0}".format(clust_id)] < 0.0
        )
        assert idx_up.sum() + idx_down.sum() == idx.sum()

        cols = [
            "{0}:{1}".format(x, clust_id)
            for x in [
                "percentage_fold_change" if de_test == "fisher" else "log_fold_change",
                "percentage",
            ]
        ]
        de_up = pd.DataFrame(
            data=data.varm[de_key][cols][idx_up], index=data.var_names[idx_up]
        )
        de_up.rename(columns={cols[0]: "fc", cols[1]: "percent"}, inplace=True)
        de_down = pd.DataFrame(
            data=data.varm[de_key][cols][idx_down], index=data.var_names[idx_down]
        )
        de_down.rename(columns={cols[0]: "fc", cols[1]: "percent"}, inplace=True)

        if de_test != "fisher":
            de_up["fc"] = np.exp(de_up["fc"])
            de_down["fc"] = np.exp(de_down["fc"])

        results = anno.evaluate(de_up, de_down, ignore_nonde=ignore_nonde)

        if output_file is not None:
            fout.write("Cluster {}:\n".format(clust_id))
            anno.report(fout, results, threshold)

        cell_type_results[clust_id] = results

    if output_file is not None:
        fout.close()

    return cell_type_results


def annotate(
    data: AnnData, name: str, based_on: str, anno_dict: Dict[str, str]
) -> None:
    """Add annotation to AnnData obj.

    Parameters
    ----------

    data : `AnnData`
        AnnData object.
    name : `str`
        Name of the new annotation in data.obs.
    based_on : `str`
        Name of the attribute the cluster ids coming from.
    anno_dict : `Dict[str, str]`
        Dictionary mapping from cluster id to cell type.

    Returns
    -------

    None

    Examples
    --------
    >>> annotate_cluster.annotate(data, 'anno', 'spectral_louvain_labels', {'1': 'T cell', '2': 'B cell'})
    """
    data.obs[name] = [anno_dict[x] for x in data.obs[based_on]]

@pg_deco.TimeLogger()
def run_annotate_cluster(
    input_file: str,
    output_file: str,
    marker_file: str,
    de_test: str,
    de_alpha: float = 0.05,
    de_key: str = "de_res",
    threshold: float = 0.5,
    ignore_nonde: bool = False,
) -> None:
    """ For command line use.
    """
    from pegasus.io import read_input

    data = read_input(input_file, h5ad_mode="r")
    infer_cell_types(
        data,
        marker_file,
        de_test,
        de_alpha=de_alpha,
        de_key=de_key,
        threshold=threshold,
        ignore_nonde=ignore_nonde,
        output_file=output_file,
    )
    data.file.close()


def annotate_anndata_object(input_file: str, annotation: str) -> None:
    """ For command line use.
        annotation:  anno_name:clust_name:cell_type1;...cell_typen
    """
    from pegasus.io import read_input, write_output

    data = read_input(input_file, h5ad_mode="r+")
    anno_name, clust_name, anno_str = annotation.split(":")
    anno_dict = {str(i + 1): x for i, x in enumerate(anno_str.split(";"))}
    annotate(data, anno_name, clust_name, anno_dict)
    write_output(data, input_file, whitelist = ["obs", "uns"])
